import logging
import time

from typing import Dict, List, Optional, Union

import orjson

from clipped.utils.json import orjson_loads
from clipped.utils.logging import LogLevels
from docker import APIClient
from docker.errors import APIError, BuildError
from urllib3.exceptions import ReadTimeoutError

from polyaxon._schemas.types import Uri
from polyaxon.exceptions import PolyaxonBuildException, PolyaxonSchemaError

_logger = logging.getLogger("polyaxon.dockerizer")


class DockerMixin:
    IS_BUILD: Optional[bool] = None

    def _prepare_log_lines(self, log_line):  # pylint:disable=too-many-branches
        raw = log_line.decode("utf-8").strip()
        raw_lines = raw.split("\n")
        log_lines = []
        status = True
        for raw_line in raw_lines:
            try:
                json_line = orjson_loads(raw_line)

                if json_line.get("error"):
                    log_lines.append(
                        "{}: {}".format(
                            LogLevels.ERROR, str(json_line.get("error", json_line))
                        )
                    )
                    status = False
                else:
                    if json_line.get("stream"):
                        log_lines.append(
                            "Building: {}".format(json_line["stream"].strip())
                        )
                    elif json_line.get("status"):
                        if not self.is_pushing and not self.IS_BUILD:
                            self.is_pushing = True
                            log_lines.append("Pushing ...")
                    elif json_line.get("aux"):
                        log_lines.append(
                            "Pushing finished: {}".format(json_line.get("aux"))
                        )
                    else:
                        log_lines.append(str(json_line))
            except orjson.JSONDecodeError:
                log_lines.append("JSON decode error: {}".format(raw_line))
        return log_lines, status

    def _handle_logs(self, log_lines):
        for log_line in log_lines:
            print(log_line)  # pylint:disable=superfluous-parens

    def _handle_log_stream(self, stream):
        log_lines = []
        status = True
        try:
            for log_line in stream:
                new_log_lines, new_status = self._prepare_log_lines(log_line)
                log_lines += new_log_lines
                if not new_status:
                    status = new_status
                self._handle_logs(log_lines)
                log_lines = []
            if log_lines:
                self._handle_logs(log_lines)
        except (BuildError, APIError) as e:
            self._handle_logs(
                [
                    "{}: Could not build the image, encountered {}".format(
                        LogLevels.ERROR, e
                    )
                ]
            )
            return False

        return status


class DockerBuilder(DockerMixin):
    IS_BUILD = True

    def __init__(
        self,
        context: str,
        destination: str,
        credstore_env: Optional[Dict] = None,
        registries: Optional[List[Union[Uri, str]]] = None,
        docker: Optional[APIClient] = None,
    ):
        from polyaxon._config.parser import ConfigParser

        get_uri = ConfigParser.parse(Uri)
        self.destination = destination

        self.context = context
        self.registries = []
        for r in registries or []:
            try:
                self.registries.append(get_uri(value=r, key="DockerBuilder"))
            except PolyaxonSchemaError:
                raise PolyaxonBuildException(
                    "Registry `{}` is not valid Uri.".format(r)
                )
        self.docker = docker or APIClient(version="auto", credstore_env=credstore_env)
        self.is_pushing = False

    def check_image(self):
        return self.docker.images(self.destination)

    def login_private_registries(self):
        if not self.registries:
            return
        for registry in self.registries:
            self.docker.login(
                username=registry.user,
                password=registry.password,
                registry=registry.host_port,
                reauth=True,
            )

    def build(self, nocache: bool = False, memory_limit: Optional[int] = None):
        limits = {
            # Disable memory swap for building
            "memswap": -1
        }
        if memory_limit:
            limits["memory"] = memory_limit

        stream = self.docker.build(
            path=self.context,
            tag=self.destination,
            forcerm=True,
            rm=True,
            pull=True,
            nocache=nocache,
            container_limits=limits,
        )
        return self._handle_log_stream(stream=stream)


class DockerPusher(DockerMixin):
    IS_BUILD = False

    def __init__(
        self,
        destination: str,
        credstore_env: Optional[Dict] = None,
        docker: Optional[APIClient] = None,
    ):
        self.destination = destination
        self.docker = docker or APIClient(version="auto", credstore_env=credstore_env)
        self.is_pushing = False

    def push(self):
        stream = self.docker.push(self.destination, stream=True)
        return self._handle_log_stream(stream=stream)


def _build(
    context: str,
    destination: str,
    nocache: bool,
    docker: Optional[APIClient] = None,
    credstore_env: Optional[Dict] = None,
    registries: Optional[List[Union[Uri, str]]] = None,
):
    """Build necessary code for a job to run"""
    _logger.info("Starting build ...")

    # Build the image
    docker_builder = DockerBuilder(
        context=context,
        destination=destination,
        credstore_env=credstore_env,
        registries=registries,
        docker=docker,
    )
    docker_builder.login_private_registries()
    if docker_builder.check_image() and not nocache:
        # Image already built
        return docker_builder
    if not docker_builder.build(nocache=nocache):
        raise PolyaxonBuildException("The docker image could not be built.")
    return docker_builder


def build(
    context: str,
    destination: str,
    nocache: bool,
    docker: Optional[APIClient] = None,
    credstore_env: Optional[Dict] = None,
    registries: Optional[List[Union[Uri, str]]] = None,
    max_retries: int = 3,
    sleep_interval: int = 1,
):
    """Build necessary code for a job to run"""
    retry = 0
    is_done = False
    while retry < max_retries and not is_done:
        if retry:
            time.sleep(sleep_interval)
        try:
            docker_builder = _build(
                context=context,
                destination=destination,
                docker=docker,
                nocache=nocache,
                credstore_env=credstore_env,
                registries=registries,
            )
            is_done = True
            return docker_builder
        except ReadTimeoutError:
            retry += 1
    if not is_done:
        raise PolyaxonBuildException(
            "The docker image could not be built, client timed out."
        )


def push(
    destination: str,
    docker: Optional[APIClient] = None,
    max_retries: int = 3,
    sleep_interval: int = 1,
):
    docker_pusher = DockerPusher(destination=destination, docker=docker)
    retry = 0
    is_done = False
    while retry < max_retries and not is_done:
        if retry:
            time.sleep(sleep_interval)
        try:
            if not docker_pusher.push():
                raise PolyaxonBuildException("The docker image could not be pushed.")
            else:
                is_done = True
        except ReadTimeoutError:
            retry += 1

    if not is_done:
        raise PolyaxonBuildException(
            "The docker image could not be pushed, client timed out."
        )


def build_and_push(
    context: str,
    destination: str,
    nocache: bool,
    credstore_env: Optional[Dict] = None,
    registries: Optional[List[Union[Uri, str]]] = None,
    max_retries: int = 3,
    sleep_interval: int = 1,
):
    """Build necessary code for a job to run and push it."""
    # Build the image
    docker_builder = build(
        context=context,
        destination=destination,
        nocache=nocache,
        credstore_env=credstore_env,
        registries=registries,
        max_retries=max_retries,
        sleep_interval=sleep_interval,
    )
    push(
        destination=destination,
        docker=docker_builder.docker,
        max_retries=max_retries,
        sleep_interval=sleep_interval,
    )



import os

from typing import List, Optional

from clipped.utils.git import (
    add_remote,
    checkout_revision,
    get_code_reference,
    git_fetch,
    git_init,
    set_remote,
    update_submodules,
)
from clipped.utils.paths import check_or_create_path
from git import Repo as GitRepo

from polyaxon._client.init import get_client_or_raise
from polyaxon._env_vars.keys import (
    ENV_KEYS_GIT_CREDENTIALS,
    ENV_KEYS_GIT_CREDENTIALS_STORE,
    ENV_KEYS_SSH_PATH,
    ENV_KEYS_SSH_PRIVATE_KEY,
)
from polyaxon._schemas.lifecycle import V1Statuses
from polyaxon.exceptions import PolyaxonContainerException
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def has_cred_access() -> bool:
    return os.environ.get(ENV_KEYS_GIT_CREDENTIALS) is not None


def has_cred_store_access() -> bool:
    cred_store_path = os.environ.get(ENV_KEYS_GIT_CREDENTIALS_STORE)
    return bool(cred_store_path and os.path.exists(cred_store_path))


def has_ssh_access() -> bool:
    ssh_path = os.environ.get(ENV_KEYS_SSH_PATH)
    return bool(ssh_path and os.path.exists(ssh_path))


def get_ssh_cmd():
    ssh_path = os.environ.get(ENV_KEYS_SSH_PATH)
    ssh_key_name = os.environ.get(ENV_KEYS_SSH_PRIVATE_KEY, "id_rsa")
    return "ssh -i {}/{} -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no".format(
        ssh_path, ssh_key_name
    )


def get_clone_url(url: str) -> str:
    if not url:
        raise ValueError(
            "Git initializer requires a valid url, received {}".format(url)
        )

    if has_cred_access():
        if "https" in url:
            _url = url.replace("https://", "")
        elif url.startswith("git@") or url.endswith(".git"):
            _url = url.replace("git@", "")
            if _url.endswith(".git"):
                _url = ".git".join(_url.split(".git")[:-1])
            _url = _url.replace(":", "/")
        else:
            _url = url
        creds = os.environ.get(ENV_KEYS_GIT_CREDENTIALS)
        # Add user:pass to the git url
        return "https://{}@{}".format(creds, _url)
    if has_cred_store_access():
        if url.startswith("git@") or url.endswith(".git"):
            _url = url.replace("git@", "")
            if _url.endswith(".git"):
                _url = ".git".join(_url.split(".git")[:-1])
            _url = _url.replace(":", "/")
            return "https://{}".format(_url)
        return url
    if has_ssh_access() and "http" in url:
        if "https" in url:
            _url = url.replace("https://", "")
        elif "http" in url:
            _url = url.replace("http://", "")
        else:
            _url = url
        parts = _url.split("/")
        _url = "{}:{}".format(parts[0], "/".join(parts[1:]))
        if _url.endswith(".git"):
            _url = ".git".join(_url.split(".git")[:-1])
        return "git@{}.git".format(_url)

    return url


def clone_git_repo(
    repo_path: str, url: str, flags: Optional[List[str]] = None
) -> GitRepo:
    if has_ssh_access():
        return GitRepo.clone_from(
            url=url,
            to_path=repo_path,
            multi_options=flags,
            env={"GIT_SSH_COMMAND": get_ssh_cmd()},
        )
    return GitRepo.clone_from(url=url, to_path=repo_path, multi_options=flags)


def clone_and_checkout_git_repo(
    repo_path: str,
    clone_url: str,
    revision: str,
    flags: Optional[List[str]] = None,
):
    clone_git_repo(repo_path=repo_path, url=clone_url, flags=flags)
    if revision:
        checkout_revision(repo_path=repo_path, revision=revision)
        if flags and "--recurse-submodules" in flags:
            update_submodules(repo_path=repo_path)


def fetch_git_repo(
    repo_path: str,
    clone_url: str,
    revision: str,
    flags: Optional[List[str]] = None,
):
    check_or_create_path(repo_path, is_dir=True)
    git_init(repo_path)
    add_remote(repo_path, clone_url)
    env = None
    if has_ssh_access():
        env = {"GIT_SSH_COMMAND": get_ssh_cmd()}
    git_fetch(repo_path=repo_path, revision=revision, flags=flags, env=env)
    if flags and "--recurse-submodules" in flags:
        update_submodules(repo_path=repo_path)


def create_code_repo(
    repo_path: str,
    url: str,
    revision: str,
    connection: Optional[str] = None,
    flags: Optional[List[str]] = None,
):
    run_client = get_client_or_raise()

    try:
        clone_url = get_clone_url(url)
    except Exception as e:
        if run_client:
            run_client.log_status(
                status=V1Statuses.WARNING,
                reason="GitInitializer",
                message="Error parsing git url. "
                "Please check the git init container's logs for more details.",
            )
        raise PolyaxonContainerException("Error parsing url: {}.".format(url)) from e

    try:
        if flags and "--experimental-fetch" in flags:
            flags.remove("--experimental-fetch")
            fetch_git_repo(
                repo_path=repo_path, clone_url=clone_url, revision=revision, flags=flags
            )
        else:
            clone_and_checkout_git_repo(
                repo_path=repo_path, clone_url=clone_url, revision=revision, flags=flags
            )
    except Exception as e:
        if run_client:
            run_client.log_status(
                status=V1Statuses.WARNING,
                reason="GitInitializer",
                message="Error cloning git repo. "
                "Please check the git init container's logs for more details.",
            )
        raise e

    # Update remote
    set_remote(repo_path=repo_path, url=url)

    if not run_client:
        return

    code_ref = get_code_reference(path=repo_path, url=url)
    artifact_run = V1RunArtifact.construct(
        name=code_ref.get("commit"),
        kind=V1ArtifactKind.CODEREF,
        connection=connection,
        summary=code_ref,
        is_input=True,
    )
    run_client.log_artifact_lineage(artifact_run)



from typing import Optional, Union
from typing_extensions import Literal

from clipped.types.ref_or_obj import RefField

from polyaxon._flow.run.base import BaseRun
from polyaxon._flow.run.dask.replica import V1DaskReplica
from polyaxon._flow.run.enums import V1RunKind
from polyaxon._flow.run.resources import V1RunResources
from polyaxon._flow.run.utils import DestinationImageMixin
from polyaxon._k8s.k8s_schemas import V1Container


class V1DaskJob(BaseRun, DestinationImageMixin):
    """Dask jobs are used to run distributed jobs using a
    [Dask cluster](https://kubernetes.dask.org/en/latest/).

    > Dask Kubernetes deploys Dask workers on Kubernetes clusters using native Kubernetes APIs.
    > It is designed to dynamically launch short-lived deployments of workers
    > during the lifetime of a job.

    The Dask job spawn a temporary adaptive Dask cluster with a
    Dask scheduler and workers to run your container.

    Args:
        kind: str, should be equal `daskjob`
        job: [V1DaskReplica](/docs/experimentation/distributed/dask-replica/), optional
        worker: [V1DaskReplica](/docs/experimentation/distributed/dask-replica/), optional
        scheduler: [V1DaskReplica](/docs/experimentation/distributed/dask-replica/), optional


    ## YAML usage

    ```yaml
    >>> run:
    >>>   kind: daskjob
    >>>   job:
    >>>   worker:
    >>>   scheduler:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Environment, V1Init, V1DaskJob, V1DaskReplica
    >>> dask_job = V1DaskJob(
    >>>     job=V1DaskReplica(...),
    >>>     worker=V1DaskReplica(...),
    >>>     scheduler=V1DaskReplica(...),
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this component's runtime is a job.

    If you are using the python client to create the runtime,
    this field is not required and is set by default.

    ```yaml
    >>> run:
    >>>   kind: daskjob
    ```

    ### job

    Dask head replica specification

    ```yaml
    >>> run:
    >>>   kind: daskjob
    >>>   job:
    >>>     container:
    >>>       image: "ghcr.io/dask/dask:latest"
    >>>       args:
    >>>       - python
    >>>       - -c
    >>>       - "from dask.distributed import Client; client = Client(); # Do some work..."
    >>>     ...
    >>>   ...
    ```

    ### worker

    List of worker replica specifications

    ```yaml
    >>> run:
    >>>   kind: daskjob
    >>>   worker:
    >>>     replicas: 2
    >>>     container:
    >>>       image: "ghcr.io/dask/dask:latest"
    >>>       args:
    >>>       - dask-worker
    >>>       - --nthreads
    >>>       - "2"
    >>>       - --name
    >>>       - $(DASK_WORKER_NAME)
    >>>       - --dashboard
    >>>       - --dashboard-address
    >>>       - "8788"
    >>>   ...
    ```

    ### scheduler

    Dask scheduler replica specification

    ```yaml
    >>> run:
    >>>   kind: daskjob
    >>>   scheduler:
    >>>     container:
    >>>       image: "ghcr.io/dask/dask:latest"
    >>>       args:
    >>>       - dask-scheduler
    >>>       - --dashboard-address
    >>>       - "8787"
    >>>   ...
    ```
    """

    _IDENTIFIER = V1RunKind.DASKJOB
    _SWAGGER_FIELDS = ["volumes", "sidecars", "container"]

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    job: Optional[Union[V1DaskReplica, RefField]]
    worker: Optional[Union[V1DaskReplica, RefField]]
    scheduler: Optional[Union[V1DaskReplica, RefField]]

    def apply_image_destination(self, image: str):
        if self.job:
            self.job.container = self.job.container or V1Container()
            self.job.container.image = image

    def get_resources(self):
        resources = V1RunResources()
        if self.job:
            resources += self.job.get_resources()
        if self.worker:
            resources += self.worker.get_resources()
        if self.scheduler:
            resources += self.scheduler.get_resources()
        return resources

    def get_all_containers(self):
        containers = []
        if self.job:
            containers += self.job.get_all_containers()
        if self.worker:
            containers += self.worker.get_all_containers()
        if self.scheduler:
            containers += self.scheduler.get_all_containers()
        return containers

    def get_all_connections(self):
        connections = []
        if self.job:
            connections += self.job.get_all_connections()
        if self.worker:
            connections += self.worker.get_all_connections()
        if self.scheduler:
            connections += self.scheduler.get_all_connections()
        return connections

    def get_all_init(self):
        init = []
        if self.job:
            init += self.job.get_all_init()
        if self.worker:
            init += self.worker.get_all_init()
        if self.scheduler:
            init += self.scheduler.get_all_init()
        return init



from typing import Optional

from clipped.compact.pydantic import Field, StrictStr
from clipped.types.ref_or_obj import IntOrRef

from polyaxon._schemas.base import BaseSchemaModel


class V1SchedulingPolicy(BaseSchemaModel):
    """Scheduling policy for Kubeflow distributed runs.

    Args:
        min_available: int, optional
        queue: str, optional
        min_resources: int, optional
        priority_class: str, optional
        schedule_timeout_seconds: int, optional
    """

    _IDENTIFIER = "schedulingPolicy"

    min_available: Optional[IntOrRef] = Field(alias="minAvailable")
    queue: Optional[StrictStr]
    min_resources: Optional[IntOrRef] = Field(alias="minResources")
    priority_class: Optional[StrictStr] = Field(alias="priorityClass")
    schedule_timeout_seconds: Optional[IntOrRef] = Field(alias="scheduleTimeoutSeconds")



from typing import List, Optional

from clipped.utils.lists import to_list

from polyaxon._auxiliaries import V1PolyaxonSidecarContainer
from polyaxon._connections import V1Connection
from polyaxon._containers.names import SIDECAR_CONTAINER
from polyaxon._env_vars.keys import ENV_KEYS_ARTIFACTS_STORE_NAME, ENV_KEYS_CONTAINER_ID
from polyaxon._flow import V1Plugins
from polyaxon._k8s import k8s_schemas
from polyaxon._runner.converter import BaseConverter as _BaseConverter
from polyaxon.exceptions import PolyaxonConverterError


class SidecarConverter(_BaseConverter):
    @classmethod
    def _get_sidecar_env_vars(
        cls,
        env_vars: List[k8s_schemas.V1EnvVar],
        container_id: str,
        artifacts_store_name: str,
    ) -> List[k8s_schemas.V1EnvVar]:
        env_vars = to_list(env_vars, check_none=True)[:]
        env_vars.append(
            cls._get_env_var(name=ENV_KEYS_CONTAINER_ID, value=container_id)
        )
        env_vars.append(
            cls._get_env_var(
                name=ENV_KEYS_ARTIFACTS_STORE_NAME, value=artifacts_store_name
            )
        )
        return env_vars

    @staticmethod
    def _get_sidecar_args(
        container_id: str,
        sleep_interval: int,
        sync_interval: int,
        monitor_logs: bool,
        monitor_spec: bool,
    ) -> List[str]:
        args = [
            "--container-id={}".format(container_id),
            "--sleep-interval={}".format(sleep_interval),
            "--sync-interval={}".format(sync_interval),
        ]
        # enable monitor logs and spec by default
        if monitor_logs is None:
            monitor_logs = True
        if monitor_spec is None:
            monitor_spec = True
        if monitor_logs:
            args.append("--monitor-logs")
        if monitor_spec:
            args.append("--monitor-spec")
        return args

    @classmethod
    def _get_sidecar_container(
        cls,
        container_id: str,
        polyaxon_sidecar: V1PolyaxonSidecarContainer,
        env: List[k8s_schemas.V1EnvVar],
        artifacts_store: V1Connection,
        plugins: V1Plugins,
        run_path: Optional[str],
    ) -> Optional[k8s_schemas.V1Container]:
        if artifacts_store and not plugins:
            raise PolyaxonConverterError(
                "Logs/artifacts store was passed and contexts was not passed."
            )

        has_artifacts = artifacts_store and plugins.collect_artifacts
        has_logs = artifacts_store and plugins.collect_logs

        if not has_logs and not has_artifacts:
            # No sidecar
            return None

        if (has_artifacts or has_logs) and not run_path:
            raise PolyaxonConverterError(
                "Logs store/outputs store must have a run_path."
            )

        env = cls._get_sidecar_env_vars(
            env_vars=env,
            container_id=container_id,
            artifacts_store_name=artifacts_store.name,
        )

        volume_mounts = cls._get_mounts(
            use_auth_context=plugins.auth,
            use_artifacts_context=has_artifacts,
            use_docker_context=False,
            use_shm_context=False,
        )

        sleep_interval = polyaxon_sidecar.sleep_interval
        sync_interval = polyaxon_sidecar.sync_interval
        monitor_logs = polyaxon_sidecar.monitor_logs
        monitor_spec = polyaxon_sidecar.monitor_spec
        if plugins and plugins.sidecar:
            if plugins.sidecar.sleep_interval:
                sleep_interval = plugins.sidecar.sleep_interval
            if plugins.sidecar.sync_interval:
                sync_interval = plugins.sidecar.sync_interval
            if plugins.sidecar.monitor_logs:
                monitor_logs = plugins.sidecar.monitor_logs
            if plugins.sidecar.monitor_spec:
                monitor_spec = plugins.sidecar.monitor_spec
        sidecar_args = cls._get_sidecar_args(
            container_id=container_id,
            sleep_interval=sleep_interval,
            sync_interval=sync_interval,
            monitor_logs=monitor_logs,
            monitor_spec=monitor_spec,
        )

        env_from = []

        if artifacts_store.is_bucket:
            secret = artifacts_store.secret
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=secret), check_none=True
            )
            env += to_list(cls._get_items_from_secret(secret=secret), check_none=True)
            env_from += to_list(
                cls._get_env_from_secret(secret=secret), check_none=True
            )

            config_map = artifacts_store.config_map
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=config_map), check_none=True
            )
            env += to_list(
                cls._get_items_from_config_map(config_map=config_map), check_none=True
            )
            env_from += to_list(
                cls._get_env_from_config_map(config_map=config_map), check_none=True
            )
        else:
            volume_mounts += to_list(
                cls._get_mount_from_store(store=artifacts_store), check_none=True
            )
        # Add connections catalog env vars information
        env += to_list(
            cls._get_connections_catalog_env_var(connections=[artifacts_store]),
            check_none=True,
        )
        env += to_list(
            cls._get_connection_env_var(connection=artifacts_store),
            check_none=True,
        )

        container = k8s_schemas.V1Container(
            name=SIDECAR_CONTAINER,
            image=polyaxon_sidecar.get_image(),
            image_pull_policy=polyaxon_sidecar.image_pull_policy,
            command=["polyaxon", "sidecar"],
            args=sidecar_args,
            env=env,
            env_from=env_from,
            resources=polyaxon_sidecar.get_resources(),
            volume_mounts=volume_mounts,
        )

        return cls._patch_container(container)



from polyaxon._sidecar.container.monitors.artifacts import sync_artifacts
from polyaxon._sidecar.container.monitors.logs import sync_logs
from polyaxon._sidecar.container.monitors.spec import sync_spec



import time
import traceback

from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, Optional, Tuple, Type

from clipped.utils.tz import now

from polyaxon import settings
from polyaxon._auxiliaries import V1PolyaxonInitContainer, V1PolyaxonSidecarContainer
from polyaxon._connections import V1Connection
from polyaxon._constants.globals import DEFAULT
from polyaxon._runner.agent.client import AgentClient
from polyaxon._runner.executor import BaseExecutor
from polyaxon._schemas.checks import ChecksConfig
from polyaxon._schemas.lifecycle import LiveState, V1Statuses
from polyaxon.client import V1AgentStateResponse
from polyaxon.exceptions import PolyaxonConverterError
from polyaxon.logger import logger


class BaseAgent:
    EXECUTOR: Type[BaseExecutor]
    HEALTH_FILE = "/tmp/.healthz"
    SLEEP_STOP_TIME = 60 * 5
    SLEEP_ARCHIVED_TIME = 60 * 60
    IS_ASYNC = False

    def __init__(
        self,
        owner: Optional[str] = None,
        agent_uuid: Optional[str] = None,
        max_interval: Optional[int] = None,
    ):
        self.max_interval = 6 if agent_uuid else 4
        if max_interval:
            self.max_interval = max(max_interval, 3)
        if not agent_uuid and not owner:
            owner = DEFAULT
        self.executor = None
        self._default_auth = bool(agent_uuid)
        self._executor_refreshed_at = now()
        self._graceful_shutdown = False
        self.client = AgentClient(
            owner=owner, agent_uuid=agent_uuid, is_async=self.IS_ASYNC
        )
        self.executor = self.EXECUTOR()
        self.content = settings.AGENT_CONFIG.to_json()

    def sync(self):
        raise NotImplementedError

    def cron(self):
        return self.client.cron_agent()

    def sync_compatible_updates(self, compatible_updates: Dict):
        if compatible_updates and settings.AGENT_CONFIG:
            init = compatible_updates.get("init")
            if init and settings.AGENT_CONFIG.init:
                init = V1PolyaxonInitContainer.from_dict(init)
                settings.AGENT_CONFIG.init = settings.AGENT_CONFIG.init.patch(init)

            sidecar = compatible_updates.get("sidecar")
            if sidecar and settings.AGENT_CONFIG.sidecar:
                sidecar = V1PolyaxonSidecarContainer.from_dict(sidecar)
                settings.AGENT_CONFIG.sidecar = settings.AGENT_CONFIG.sidecar.patch(
                    sidecar
                )
            connections = compatible_updates.get("connections")
            if connections:
                settings.AGENT_CONFIG.connections = [
                    V1Connection.from_dict(c) for c in connections
                ]

            self.content = settings.AGENT_CONFIG.to_json()
            return self.sync()

    @classmethod
    def get_healthz_config(cls) -> Optional[ChecksConfig]:
        try:
            return ChecksConfig.read(cls.HEALTH_FILE, config_type=".json")
        except Exception:  # noqa
            return None

    @classmethod
    def ping(cls):
        if not settings.AGENT_CONFIG.enable_health_checks:
            return
        ChecksConfig.init_file(cls.HEALTH_FILE)
        config = cls.get_healthz_config()
        if config:
            config.last_check = now()
            config.write(cls.HEALTH_FILE, mode=config._WRITE_MODE)

    @classmethod
    def pong(cls, interval: int = 15) -> bool:
        config = cls.get_healthz_config()
        if not config:
            return False
        return not config.should_check(interval=interval)

    def refresh_executor(self):
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def end(self, sleep: Optional[int] = None):
        self._graceful_shutdown = True
        if sleep:
            time.sleep(sleep)
        else:
            logger.info("Agent is shutting down.")

    def _check_status(self, agent_state):
        if agent_state.status == V1Statuses.STOPPED:
            print(
                "Agent has been stopped from the platform,"
                "but the deployment is still running."
                "Please either set the agent to starting or teardown the agent deployment."
            )
            return self.end(sleep=self.SLEEP_STOP_TIME)
        elif agent_state.live_state < LiveState.LIVE:
            print(
                "Agent has been archived from the platform,"
                "but the deployment is still running."
                "Please either restore the agent or teardown the agent deployment."
            )
            return self.end(sleep=self.SLEEP_ARCHIVED_TIME)

    def process(self, pool: "ThreadPoolExecutor") -> V1AgentStateResponse:
        raise NotImplementedError

    def make_run_resource(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
        default_auth: bool = False,
    ) -> Optional[Any]:
        try:
            return self.executor.make_and_convert(
                owner_name=owner_name,
                project_name=project_name,
                run_uuid=run_uuid,
                run_name=run_name,
                content=content,
                default_auth=default_auth,
            )
        except PolyaxonConverterError as e:
            logger.info(
                "Run could not be cleaned. Agent failed converting run manifest: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )
        except Exception as e:
            logger.info(
                "Agent failed during compilation with unknown exception: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )
        return None

    def prepare_run_resource(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
    ) -> Optional[Any]:
        raise NotImplementedError

    def submit_run(self, run_data: Tuple[str, str, str, str, str]):
        raise NotImplementedError

    def make_and_create_run(
        self, run_data: Tuple[str, str, str, str, str], default_auth: bool = False
    ):
        raise NotImplementedError

    def apply_run(self, run_data: Tuple[str, str, str, str, str]):
        raise NotImplementedError

    def check_run(self, run_data: Tuple[str, str, str]):
        raise NotImplementedError

    def stop_run(self, run_data: Tuple[str, str, str]):
        raise NotImplementedError

    def delete_run(self, run_data: Tuple[str, str, str, str, str]):
        raise NotImplementedError

    def clean_run(self, run_uuid: str, run_kind: str, namespace: str = None):
        raise NotImplementedError



from typing import Dict, List, Optional

from polyaxon._flow import V1Notification, V1Termination
from polyaxon._k8s import k8s_schemas
from polyaxon._k8s.converter.pod.spec import get_pod_spec, get_pod_template_spec
from polyaxon._k8s.custom_resources.operation import get_operation_custom_object
from polyaxon._k8s.custom_resources.setter import (
    set_collect_logs,
    set_notify,
    set_sync_statuses,
    set_termination,
)
from polyaxon._k8s.replica import ReplicaSpec


def _get_ray_replicas_template(
    namespace: str,
    resource_name: str,
    replica_name: Optional[str],
    replica: Optional[ReplicaSpec],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
    default_start_params: Optional[Dict] = None,
):
    if not replica:
        return

    metadata, pod_spec = get_pod_spec(
        namespace=namespace,
        main_container=replica.main_container,
        sidecar_containers=replica.sidecar_containers,
        init_containers=replica.init_containers,
        resource_name=resource_name,
        volumes=replica.volumes,
        environment=replica.environment,
        labels=labels,
        annotations=annotations,
    )

    data = {
        "replicas": replica.num_replicas,
        "restartPolicy": pod_spec.restart_policy or "Never",
        "template": get_pod_template_spec(metadata=metadata, pod_spec=pod_spec),
    }
    custom = replica.custom or {}
    if custom.get("min_replicas"):
        data["minReplicas"] = custom["min_replicas"]
    if custom.get("max_replicas"):
        data["maxReplicas"] = custom["max_replicas"]
    if custom.get("ray_start_params"):
        data["rayStartParams"] = custom["ray_start_params"]
    if default_start_params:
        data["rayStartParams"] = {
            **default_start_params,
            **data.get("rayStartParams", {}),
        }
    if replica_name:
        data["groupName"] = replica_name
    return data


def get_ray_head_replicas_template(
    namespace: str,
    resource_name: str,
    replica: Optional[ReplicaSpec],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
):
    if replica:
        # Set default ports on main container
        if not replica.main_container.ports:
            replica.main_container.ports = [
                k8s_schemas.V1ContainerPort(
                    container_port=6379, name="gcs-server", protocol="TCP"
                ),
                k8s_schemas.V1ContainerPort(
                    container_port=8265, name="dashboard", protocol="TCP"
                ),
                k8s_schemas.V1ContainerPort(
                    container_port=10001, name="client", protocol="TCP"
                ),
                k8s_schemas.V1ContainerPort(
                    container_port=8000, name="serve", protocol="TCP"
                ),
            ]
    head = _get_ray_replicas_template(
        namespace=namespace,
        resource_name=resource_name,
        replica_name="head",
        replica=replica,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
        default_start_params={"dashboard-host": "0.0.0.0"},
    )
    if head:
        template_spec["head"] = head


def get_ray_worker_replicas_template(
    namespace: str,
    resource_name: str,
    replicas: Optional[Dict[str, ReplicaSpec]],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
):
    workers = []
    for replica_name in replicas or {}:
        replica = replicas[replica_name]
        worker = _get_ray_replicas_template(
            namespace=namespace,
            resource_name=resource_name,
            replica_name=replica_name,
            replica=replica,
            labels=labels,
            annotations=annotations,
            template_spec=template_spec,
        )
        # Check the lifecycle is set
        if worker:
            template = worker["template"]
            if template.spec.containers[0].lifecycle is None:
                template.spec.containers[0].lifecycle = {
                    "preStop": {"exec": {"command": ["/bin/sh", "-c", "ray stop"]}}
                }
                worker["template"] = template
            workers.append(worker)
    if workers:
        template_spec["workers"] = workers


def get_ray_job_custom_resource(
    resource_name: str,
    namespace: str,
    head: Optional[ReplicaSpec],
    workers: Optional[Dict[str, ReplicaSpec]],
    termination: V1Termination,
    collect_logs: bool,
    sync_statuses: bool,
    notifications: List[V1Notification],
    entrypoint: Optional[str],
    metadata: Optional[Dict[str, str]],
    runtime_env: Optional[str],
    ray_version: Optional[str],
    labels: Dict[str, str],
    annotations: Dict[str, str],
) -> Dict:
    template_spec = {}

    get_ray_head_replicas_template(
        replica=head,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
    )
    get_ray_worker_replicas_template(
        replicas=workers,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
    )
    if entrypoint:
        template_spec["entrypoint"] = entrypoint
    if metadata:
        template_spec["metadata"] = metadata
    if runtime_env:
        template_spec["runtimeEnv"] = runtime_env
    if ray_version:
        template_spec["rayVersion"] = ray_version

    custom_object = {"rayJobSpec": template_spec}
    custom_object = set_termination(
        custom_object=custom_object, termination=termination
    )
    custom_object = set_collect_logs(
        custom_object=custom_object, collect_logs=collect_logs
    )
    custom_object = set_sync_statuses(
        custom_object=custom_object, sync_statuses=sync_statuses
    )
    custom_object = set_notify(custom_object=custom_object, notifications=notifications)

    return get_operation_custom_object(
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        custom_object=custom_object,
    )



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1Installation(BaseAllowSchemaModel):
    key: Optional[StrictStr]
    version: Optional[StrictStr]
    dist: Optional[StrictStr]
    host: Optional[StrictStr]
    hmac: Optional[StrictStr]
    mode: Optional[StrictStr]
    org: Optional[bool]
    auth: Optional[List[StrictStr]]



from typing import Optional

LOCAL = "local"
SANDBOX = "sandbox"
CLOUD = "cloud"
CE = "ce"
EE = "ee"


def is_local(value: str):
    return LOCAL == value


def is_sandbox(value: str):
    return SANDBOX == value


def is_ce(value: str):
    return CE == value


def is_community(value: str):
    return value is None or value in {LOCAL, SANDBOX, CE}


def is_cloud(value: str):
    return CLOUD == value


def is_ee(value: str):
    return EE == value


def get_dist(value: str, default: Optional[str] = None):
    if value and value in {LOCAL, SANDBOX, CE, CLOUD, EE}:
        return value
    return default



import json
import subprocess

from tempfile import TemporaryFile

from polyaxon.exceptions import PolyaxonOperatorException


class CmdOperator:
    CMD = ""

    @classmethod
    def _execute(
        cls,
        params,
        env,
        is_json: bool = False,
        stream: bool = False,
        output_only: bool = True,
    ):
        def _stream():
            with TemporaryFile("w+") as stderr:
                ps = subprocess.Popen(params, env=env, stderr=stderr)
                exit_status = ps.wait()
                stderr.seek(0)
                if exit_status != 0:
                    raise PolyaxonOperatorException(
                        cmd=cls.CMD,
                        args=params,
                        return_code=exit_status,
                        stdout=None,
                        stderr=stderr,
                    )

        def _block():
            with TemporaryFile("w+") as stdout, TemporaryFile("w+") as stderr:
                ps = subprocess.Popen(params, env=env, stdout=stdout, stderr=stderr)
                exit_status = ps.wait()
                stdout.seek(0)
                stderr.seek(0)
                if exit_status != 0:
                    raise PolyaxonOperatorException(
                        cmd=cls.CMD,
                        args=params,
                        return_code=exit_status,
                        stdout=stdout,
                        stderr=stderr,
                    )

                return json.load(stdout) if is_json else stdout.read()

        if output_only:
            return _stream() if stream else _block()

        return subprocess.Popen(params, env=env, stdout=subprocess.PIPE)

    @classmethod
    def check(cls):
        return True



from typing import Any, Dict, List, Optional

from clipped.compact.pydantic import Field, StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1SectionSpec(BaseAllowSchemaModel):
    name: Optional[StrictStr]
    is_minimized: Optional[bool]
    columns: Optional[int]
    height: Optional[int]
    widgets: Optional[List[Dict[str, Any]]]
    page_index: Optional[int] = Field(alias="pageIndex")
    page_size: Optional[int] = Field(alias="pageSize")



from polyaxon._schemas.base import BaseSchemaModel


class BaseRun(BaseSchemaModel):
    _USE_DISCRIMINATOR = True

    def get_resources(self):
        raise NotImplementedError

    def get_all_containers(self):
        raise NotImplementedError

    def get_all_connections(self):
        raise NotImplementedError

    def get_all_init(self):
        raise NotImplementedError



import os

from typing import List, Optional

from clipped.utils.paths import create_tarfile, get_files_in_path

from polyaxon import settings


def tar_dir(download_path: str) -> str:
    outputs_files = get_files_in_path(download_path)
    tar_base_name = os.path.basename(download_path)
    tar_name = "{}.tar.gz".format(tar_base_name)
    target_path = os.path.join(settings.CLIENT_CONFIG.archives_root or "", tar_name)
    create_tarfile(files=outputs_files, tar_path=target_path, relative_to=download_path)
    return target_path


def tar_files(
    download_name: str, outputs_files: List[str], relative_to: Optional[str] = None
) -> str:
    tar_name = "{}.tar.gz".format(download_name)
    target_path = os.path.join(settings.CLIENT_CONFIG.archives_root or "", tar_name)
    create_tarfile(files=outputs_files, tar_path=target_path, relative_to=relative_to)
    return target_path



from typing import Optional

from clipped.utils.lists import to_list

from polyaxon._containers.names import generate_container_name
from polyaxon._runner.converter.types import Container


def sanitize_container_command_args(
    container: Container,
) -> Container:
    # Sanitize container command/args
    if container.command:
        container.command = [
            str(c) for c in to_list(container.command, check_none=True) if c
        ]
    if container.args:
        container.args = [str(c) for c in to_list(container.args, check_none=True) if c]

    return container


def ensure_container_name(
    container: Container, prefix: Optional[str] = None
) -> Container:
    if not container:
        return container

    name = container.name
    if not name:
        container.name = generate_container_name(prefix=prefix)
    return container



import os

from typing import Optional

from polyaxon._utils.fqn_utils import to_fqn_name
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def collect_lineage_artifacts_path(artifact_path: str) -> Optional[V1RunArtifact]:
    name = os.path.basename(
        artifact_path.rstrip("/")
    )  # Trim handles cases like `foo/` -> ''
    return V1RunArtifact(
        name=to_fqn_name(name) if name else "_",
        kind=V1ArtifactKind.DIR,
        path=artifact_path,
        summary={"path": artifact_path},
        is_input=True,
    )



import datetime

from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._sdk.schemas.v1_dashboard_spec import V1DashboardSpec


class V1Dashboard(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    live_state: Optional[int]
    spec: Optional[V1DashboardSpec]
    org_level: Optional[bool]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]



from typing import Union
from typing_extensions import Annotated

from clipped.compact.pydantic import Field

from polyaxon._flow.matrix.bayes import (
    GaussianProcessConfig,
    UtilityFunctionConfig,
    V1Bayes,
)
from polyaxon._flow.matrix.enums import (
    AcquisitionFunctions,
    GaussianProcessesKernels,
    V1HPKind,
    V1MatrixKind,
)
from polyaxon._flow.matrix.grid_search import V1GridSearch
from polyaxon._flow.matrix.hyperband import V1Hyperband
from polyaxon._flow.matrix.hyperopt import V1Hyperopt
from polyaxon._flow.matrix.iterative import V1Iterative
from polyaxon._flow.matrix.mapping import V1Mapping
from polyaxon._flow.matrix.params import (
    V1HpChoice,
    V1HpDateRange,
    V1HpDateTimeRange,
    V1HpGeomSpace,
    V1HpLinSpace,
    V1HpLogNormal,
    V1HpLogSpace,
    V1HpLogUniform,
    V1HpNormal,
    V1HpParam,
    V1HpPChoice,
    V1HpQLogNormal,
    V1HpQLogUniform,
    V1HpQNormal,
    V1HpQUniform,
    V1HpRange,
    V1HpUniform,
    validate_pchoice,
)
from polyaxon._flow.matrix.random_search import V1RandomSearch
from polyaxon._flow.matrix.tuner import V1Tuner

V1Matrix = Annotated[
    Union[
        V1Bayes,
        V1GridSearch,
        V1Hyperband,
        V1Hyperopt,
        V1Iterative,
        V1Mapping,
        V1RandomSearch,
    ],
    Field(discriminator="kind"),
]


class MatrixMixin:
    def get_matrix_kind(self):
        raise NotImplementedError

    @property
    def has_mapping_matrix(self):
        return self.get_matrix_kind() == V1Mapping._IDENTIFIER

    @property
    def has_grid_search_matrix(self):
        return self.get_matrix_kind() == V1GridSearch._IDENTIFIER

    @property
    def has_random_search_matrix(self):
        return self.get_matrix_kind() == V1RandomSearch._IDENTIFIER

    @property
    def has_hyperband_matrix(self):
        return self.get_matrix_kind() == V1Hyperband._IDENTIFIER

    @property
    def has_bo_matrix(self):
        return self.get_matrix_kind() == V1Bayes._IDENTIFIER

    @property
    def has_hyperopt_matrix(self):
        return self.get_matrix_kind() == V1Hyperopt._IDENTIFIER

    @property
    def has_iterative_matrix(self):
        return self.get_matrix_kind() == V1Iterative._IDENTIFIER



from typing import Dict, Iterable, List, Optional, Union

from polyaxon import settings
from polyaxon._auxiliaries import V1PolyaxonSidecarContainer
from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._docker import docker_types
from polyaxon._docker.converter.base.containers import ContainerMixin
from polyaxon._docker.converter.base.env_vars import EnvMixin
from polyaxon._docker.converter.base.init import InitConverter
from polyaxon._docker.converter.base.main import MainConverter
from polyaxon._docker.converter.base.mounts import MountsMixin
from polyaxon._flow import V1Environment, V1Init, V1Plugins
from polyaxon._k8s import k8s_schemas
from polyaxon._runner.converter import BaseConverter as _BaseConverter
from polyaxon._runner.kinds import RunnerKind
from polyaxon.exceptions import PolyaxonConverterError


class BaseConverter(
    MainConverter, InitConverter, ContainerMixin, EnvMixin, MountsMixin, _BaseConverter
):
    RUNNER_KIND = RunnerKind.DOCKER

    @staticmethod
    def _post_process_host(host: str) -> str:
        for v in ["0.0.0.0", "127.0.0.1", "localhost"]:
            if v in host:
                return host.replace(v, "host.docker.internal")
        return host

    @classmethod
    def _get_sidecar_container(
        cls,
        container_id: str,
        polyaxon_sidecar: V1PolyaxonSidecarContainer,
        env: List[docker_types.V1EnvVar],
        artifacts_store: V1Connection,
        plugins: V1Plugins,
        run_path: Optional[str],
    ) -> Optional[docker_types.V1Container]:
        return None

    @classmethod
    def _k8s_to_docker_env_var(
        cls,
        env_var: List[k8s_schemas.V1EnvVar],
    ) -> List[docker_types.V1EnvVar]:
        if not env_var:
            return []

        docker_env_var = []
        for item in env_var:
            if isinstance(item, dict):
                try:
                    item = k8s_schemas.V1EnvVar(**item)
                except (ValueError, TypeError) as e:
                    raise PolyaxonConverterError(
                        f"Could not parse env var value `{item}` from the K8S schema in container section"
                    ) from e
            docker_env_var.append(cls._get_env_var(name=item.name, value=item.value))

        return docker_env_var

    @classmethod
    def _k8s_to_docker_volume_mounts(
        cls,
        volume_mounts: List[k8s_schemas.V1VolumeMount],
        volumes: List[k8s_schemas.V1Volume],
    ) -> List[docker_types.V1VolumeMount]:
        if not volume_mounts or not volumes:
            return []

        _volume_mounts = []
        for item in volume_mounts:
            if isinstance(item, dict):
                try:
                    _volume_mounts.append(k8s_schemas.V1VolumeMount(item))
                except (ValueError, TypeError) as e:
                    raise PolyaxonConverterError(
                        f"Could not parse volume mount value `{item}` from the K8S schema in container section"
                    ) from e
            else:
                _volume_mounts.append(item)
        volume_mounts = _volume_mounts

        _volumes = []
        for item in volumes:
            if isinstance(item, dict):
                try:
                    _volumes.append(k8s_schemas.V1Volume(**item))
                except (ValueError, TypeError) as e:
                    raise PolyaxonConverterError(
                        f"Could not parse volume value `{item}` from the K8S schema in container section"
                    ) from e
        volumes = _volumes

        docker_volume_mounts = []
        volumes_by_name = {item.name: item for item in volumes}
        for item in volume_mounts:
            volume = volumes_by_name.get(item.name)  # type: k8s_schemas.V1Volume
            if not volume:
                continue
            if volume.host_path:
                host_path = volume.host_path.path
            elif volume.persistent_volume_claim:
                host_path = volume.persistent_volume_claim.claim_name
            else:
                host_path = None
            docker_volume_mounts.append(
                cls._get_volume(
                    mount_path=item.mount_path,
                    host_path=host_path,
                    read_only=item.read_only,
                )
            )

        return docker_volume_mounts

    @staticmethod
    def _k8s_to_docker_resources(
        resources: k8s_schemas.V1ResourceRequirements,
    ) -> Optional[docker_types.V1ResourceRequirements]:
        if not resources:
            return None

        if isinstance(resources, dict):
            try:
                resources = k8s_schemas.V1ResourceRequirements(**resources)
            except (ValueError, TypeError) as e:
                raise PolyaxonConverterError(
                    f"Could not parse resources value `{resources}` from the K8S schema in container section"
                ) from e

        cpus = None
        memory = None
        gpus = None
        if resources.limits:
            cpus = resources.limits.get("cpu")
            if not cpus:
                cpus = resources.limits.get("cpus")
            gpus = resources.limits.get("gpu")
            if not gpus:
                gpus = resources.limits.get("gpus")
            if not gpus:
                gpus = resources.limits.get("nvidia.com/gpu")
            memory = resources.limits.get("memory")
        if resources.requests:
            cpus = cpus or resources.requests.get("cpu")
            if not cpus:
                cpus = resources.requests.get("cpus")
            gpus = gpus or resources.requests.get("gpu")
            if not gpus:
                gpus = resources.requests.get("gpus")
            if not gpus:
                gpus = resources.requests.get("nvidia.com/gpu")
            memory = memory or resources.requests.get("memory")
        docker_resources = {}
        if cpus:
            docker_resources["cpus"] = docker_types.V1ResourceRequirements.from_k8s_cpu(
                cpus
            )
        if memory:
            docker_resources[
                "memory"
            ] = docker_types.V1ResourceRequirements.from_k8s_memory(memory)
        if gpus:
            docker_resources["gpus"] = gpus
        return docker_types.V1ResourceRequirements.from_dict(docker_resources)

    @staticmethod
    def _new_container(name: str) -> docker_types.V1Container:
        return docker_types.V1Container(name=name)

    @classmethod
    def _ensure_container(
        cls,
        container: Union[k8s_schemas.V1Container, docker_types.V1Container],
        volumes: List[k8s_schemas.V1Volume],
    ) -> docker_types.V1Container:
        if not isinstance(container, k8s_schemas.V1Container):
            return container
        return docker_types.V1Container(
            name=container.name,
            image=container.image,
            command=container.command,
            args=container.args,
            env=cls._k8s_to_docker_env_var(container.env),
            volume_mounts=cls._k8s_to_docker_volume_mounts(
                container.volume_mounts, volumes
            ),
            resources=cls._k8s_to_docker_resources(container.resources),
            working_dir=container.working_dir,
        )

    def get_replica_resource(
        self,
        environment: V1Environment,
        plugins: V1Plugins,
        volumes: List[k8s_schemas.V1Volume],
        init: List[V1Init],
        sidecars: List[k8s_schemas.V1Container],
        container: k8s_schemas.V1Container,
        artifacts_store: V1Connection,
        connections: List[str],
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        kv_env_vars: List[List],
        default_sa: Optional[str] = None,
        ports: List[int] = None,
    ) -> List[docker_types.V1Container]:
        volumes = volumes or []
        init = init or []
        sidecars = sidecars or []
        connections = connections or []
        environment = environment or V1Environment()
        environment.service_account_name = (
            environment.service_account_name
            or default_sa
            or settings.AGENT_CONFIG.runs_sa
        )

        init_connections = self.filter_connections_from_init(init=init)

        init_containers = self.get_init_containers(
            polyaxon_init=self.polyaxon_init,
            plugins=plugins,
            artifacts_store=artifacts_store,
            init_connections=init_connections,
            init_containers=self.filter_containers_from_init(init=init),
            connection_by_names=connection_by_names,
            log_level=plugins.log_level,
            volumes=volumes,
        )

        sidecar_containers = self.get_sidecar_containers(
            polyaxon_sidecar=self.polyaxon_sidecar,
            plugins=plugins,
            artifacts_store=artifacts_store,
            sidecar_containers=sidecars,
            log_level=plugins.log_level,
            volumes=volumes,
        )

        main_container = self.get_main_container(
            main_container=self._ensure_container(container, volumes=volumes),
            plugins=plugins,
            artifacts_store=artifacts_store,
            connections=connections,
            init_connections=init_connections,
            connection_by_names=connection_by_names,
            secrets=secrets,
            config_maps=config_maps,
            ports=ports,
            kv_env_vars=kv_env_vars,
        )

        return init_containers + sidecar_containers + [main_container]



from typing import Dict, List, Optional, Set, Union

from clipped.compact.pydantic import Field, StrictStr
from clipped.types.ref_or_obj import RefField

from polyaxon._contexts import refs as ctx_refs
from polyaxon._flow.component.base import BaseComponent
from polyaxon._flow.events import V1EventKind, V1EventTrigger
from polyaxon._flow.joins import V1Join
from polyaxon._flow.matrix import MatrixMixin, V1Matrix
from polyaxon._flow.schedules import ScheduleMixin, V1Schedule
from polyaxon._flow.trigger_policies import V1TriggerPolicy
from polyaxon._schemas.lifecycle import V1Statuses


class BaseOp(BaseComponent, MatrixMixin, ScheduleMixin):
    _FIELDS_SAME_KIND_PATCH = ["schedule", "matrix"]

    schedule: Optional[V1Schedule]
    events: Optional[Union[List[V1EventTrigger], RefField]]
    matrix: Optional[V1Matrix]
    joins: Optional[Union[List[V1Join], RefField]]
    dependencies: Optional[Union[List[StrictStr], RefField]]
    trigger: Optional[Union[V1TriggerPolicy, RefField]]
    conditions: Optional[StrictStr]
    skip_on_upstream_skip: Optional[bool] = Field(alias="skipOnUpstreamSkip")

    def get_matrix_kind(self):
        return self.matrix.kind if self.matrix else None

    def get_schedule_kind(self):
        return self.schedule.kind if self.schedule else None

    def get_upstream_statuses_events(self, upstream: Set) -> Dict[str, V1Statuses]:
        statuses_by_refs = {u: [] for u in upstream}
        events = self.events or []  # type: List[V1EventTrigger]
        for e in events:
            entity_ref = ctx_refs.get_entity_ref(e.ref)
            if not entity_ref:
                continue
            if entity_ref not in statuses_by_refs:
                continue
            for kind in e.kinds:
                status = V1EventKind.get_events_statuses_mapping().get(kind)
                if status:
                    statuses_by_refs[entity_ref].append(status)

        return statuses_by_refs

    def has_events_for_upstream(self, upstream: str) -> bool:
        events = self.events or []  # type: List[V1EventTrigger]
        for e in events:
            entity_ref = ctx_refs.get_entity_ref(e.ref)
            if not entity_ref:
                continue
            if entity_ref == upstream:
                return True

        return False



from typing import List, Optional

from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_section_spec import V1SectionSpec


class V1DashboardSpec(BaseAllowSchemaModel):
    sections: Optional[List[V1SectionSpec]]



import io
import logging
import re
import ssl

from urllib.parse import urlencode

import aiohttp

from clipped.utils.json import orjson_dumps

from polyaxon.exceptions import ApiException, ApiValueError

logger = logging.getLogger(__name__)


class RESTResponse(io.IOBase):
    def __init__(self, resp, data):
        self.aiohttp_response = resp
        self.status = resp.status
        self.reason = resp.reason
        self.data = data

    def getheaders(self):
        """Returns a CIMultiDictProxy of the response headers."""
        return self.aiohttp_response.headers

    def getheader(self, name, default=None):
        """Returns a given response header."""
        return self.aiohttp_response.headers.get(name, default)


class RESTClientObject(object):
    def __init__(self, configuration, pools_size=4, maxsize=None):
        # maxsize is number of requests to host that are allowed in parallel
        if maxsize is None:
            maxsize = configuration.connection_pool_maxsize

        ssl_context = ssl.create_default_context(cafile=configuration.ssl_ca_cert)
        if configuration.cert_file:
            ssl_context.load_cert_chain(
                configuration.cert_file, keyfile=configuration.key_file
            )

        if not configuration.verify_ssl:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

        connector = aiohttp.TCPConnector(limit=maxsize, ssl=ssl_context)

        self.proxy = configuration.proxy
        self.proxy_headers = configuration.proxy_headers

        # https pool manager
        self.pool_manager = aiohttp.ClientSession(
            connector=connector,
            trust_env=True,
            # Watch events containing large resource objects can exceed
            # aiohttp's default read buffer size.
            #
            # There is no hard-limit defined by k8s, but the etcd default
            # maximum request size is 1.5MiB.
            # https://github.com/kubernetes/kubernetes/issues/19781
            read_bufsize=2**21,
        )

    async def close(self):
        await self.pool_manager.close()

    async def request(
        self,
        method,
        url,
        query_params=None,
        headers=None,
        body=None,
        post_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        """Execute request

        :param method: http request method
        :param url: http request url
        :param query_params: query parameters in the url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _preload_content: this is a non-applicable field for
                                 the AiohttpClient.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        if post_params and body:
            raise ApiValueError(
                "body parameter cannot be used with post_params parameter."
            )

        post_params = post_params or {}
        headers = headers or {}
        # url already contains the URL query string
        # so reset query_params to empty dict
        query_params = {}
        timeout = _request_timeout or 5 * 60

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        args = {"method": method, "url": url, "timeout": timeout, "headers": headers}

        if self.proxy:
            args["proxy"] = self.proxy
        if self.proxy_headers:
            args["proxy_headers"] = self.proxy_headers

        if query_params:
            args["url"] += "?" + urlencode(query_params)

        # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
        if method in ["POST", "PUT", "PATCH", "OPTIONS", "DELETE"]:
            if re.search("json", headers["Content-Type"], re.IGNORECASE):
                if body is not None:
                    body = orjson_dumps(body)
                args["data"] = body
            elif (
                headers["Content-Type"] == "application/x-www-form-urlencoded"
            ):  # noqa: E501
                args["data"] = aiohttp.FormData(post_params)
            elif headers["Content-Type"] == "multipart/form-data":
                # must del headers['Content-Type'], or the correct
                # Content-Type which generated by aiohttp
                del headers["Content-Type"]
                data = aiohttp.FormData()
                for param in post_params:
                    k, v = param
                    if isinstance(v, tuple) and len(v) == 3:
                        data.add_field(k, value=v[1], filename=v[0], content_type=v[2])
                    else:
                        data.add_field(k, v)
                args["data"] = data

            # Pass a `bytes` parameter directly in the body to support
            # other content types than Json when `body` argument is provided
            # in serialized form
            elif isinstance(body, bytes):
                args["data"] = body
            else:
                # Cannot generate the request from given parameters
                msg = """Cannot prepare a request message for provided
                         arguments. Please check that your arguments match
                         declared content type."""
                raise ApiException(status=0, reason=msg)

        r = await self.pool_manager.request(**args)
        if _preload_content:
            data = await r.read()
            r = RESTResponse(r, data)

            # log response body
            logger.debug("response body: %s", r.data)

            if not 200 <= r.status <= 299:
                raise ApiException(http_resp=r)

        return r

    async def get_request(
        self,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "GET",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    async def head_request(
        self,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "HEAD",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    async def options_request(
        self,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "OPTIONS",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    async def delete_request(
        self,
        url,
        headers=None,
        query_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "DELETE",
            url,
            headers=headers,
            query_params=query_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    async def post_request(
        self,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "POST",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    async def put_request(
        self,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "PUT",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    async def patch_request(
        self,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return await self.request(
            "PATCH",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )



from typing import List, Optional

from clipped.compact.pydantic import StrictStr

from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._schemas.lifecycle import V1Statuses


class V1Notification(BaseSchemaModel):
    """You can configure Polyaxon to send notifications to users and systems
    about event changes in your runs.

    Notifications also allow you to build or set up integrations using webhooks.
    External systems can subscribe or provide handling for certain events.
    When one of those events is triggered, Polyaxon will send an HTTP request
    payload to the webhook's configured URL.

    Polyaxon can send notifications when a run reaches a final status:

     * succeeded
     * failed
     * stopped
     * done (any final state)

     Args:
         connections: List[str]
         trigger: str

    ## YAML usage

    ```yaml
    >>> notification:
    >>>   connections: [slack-connection, discord-connection]
    >>>   trigger: failed
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Statuses,  V1Notification
    >>> notification = V1Notification(
    >>>     connections=["slack-connection", "discord-connection"],
    >>>     trigger=V1Statuses.FAILED,
    >>> )
    ```

    ## Fields

    ### connections

    The connections to notify, these [connections](/docs/setup/connections/)
    must be configured at deployment time.

    ```yaml
    >>> notification:
    >>>   connections: [slack-connection, discord-connection]
    ```

    ### trigger

    The trigger represents the status condition to check before sending the notification.

    ```yaml
    >>> notification:
    >>>   trigger: succeeded
    ```

    In this example, the notification will be sent if the run succeeds.
    """

    _IDENTIFIER = "notification"

    connections: List[StrictStr]
    trigger: Optional[V1Statuses]

    def to_operator(self):
        data = super().to_dict()
        data["trigger"] = self.trigger.capitalize()
        return data



import os

import click

from clipped.formatting import Printer
from clipped.utils.bools import to_bool

from polyaxon import settings
from polyaxon._cli.admin import admin
from polyaxon._cli.artifacts import artifacts
from polyaxon._cli.auth import login, logout, whoami
from polyaxon._cli.check import check
from polyaxon._cli.completion import completion
from polyaxon._cli.components import components
from polyaxon._cli.config import config
from polyaxon._cli.dashboard import dashboard
from polyaxon._cli.init import init
from polyaxon._cli.models import models
from polyaxon._cli.operations import ops
from polyaxon._cli.port_forward import port_forward
from polyaxon._cli.projects import project
from polyaxon._cli.run import run
from polyaxon._cli.session import set_versions_config
from polyaxon._cli.version import check_cli_version, upgrade, version
from polyaxon._services.values import PolyaxonServices
from polyaxon.logger import clean_outputs, configure_logger

DOCS_GEN = to_bool(os.environ.get("POLYAXON_DOCS_GEN", False))


@click.group()
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Turn on debug logging"
)
@click.option(
    "--offline",
    is_flag=True,
    default=False,
    help="Run command in offline mode if supported. "
    "Currently used for run command in --local mode.",
)
@click.pass_context
@clean_outputs
def cli(context, verbose, offline):
    """Polyaxon - Cloud Native Machine Learning Automation & Experimentation tool.

    This CLI provides tools to:

      - Parse, Validate, and Check Polyaxonfiles.

      - Interact with Polyaxon server.

      - Run and Monitor experiments and jobs.

    This CLI tool comes with a caching mechanism:

      - You can initialize a project with: polyaxon init -p [project name]

      - Otherwise Polyaxon will use the default global path for the cache.

    You can check the version of your CLI by running:

      - polyaxon version

    You can check the version of the CLI and the server, and the compatibility matrix with:

      - polyaxon version --check

    To enable the debug mode, you can use the `-v` flag:

      - polyaxon -v ...

    To configure your host:

      - polyaxon config set --host=...

    To check your current config:

      - polyaxon config show

    Common commands:

      - polyaxon project get

      - polyaxon run [-f] [-l]

      - polyaxon ops ls

      - polyaxon ops logs

      - polyaxon ops get

      - polyaxon config set ...

    Admin deployment commands:

      - polyaxon admin deploy [-f] [--check]

      - polyaxon admin upgrade [-f] [--check]

      - polyaxon admin teardown [-f]

    For more information, please visit https://polyaxon.com/docs/core/cli/

    Check the help available for each command listed below by appending `-h`.
    """
    settings.set_cli_config()
    configure_logger(verbose)
    if settings.CLIENT_CONFIG.no_op:
        Printer.warning(
            "POLYAXON_NO_OP is set to `true`, some commands will not function correctly."
        )
    context.obj = context.obj or {}
    if not settings.CLIENT_CONFIG.client_header:
        settings.CLIENT_CONFIG.set_cli_header()
    context.obj["offline"] = offline
    if offline:
        os.environ["POLYAXON_IS_OFFLINE"] = "true"
        settings.CLIENT_CONFIG.is_offline = True
    non_check_cmds = [
        "completion",
        "config",
        "version",
        "login",
        "logout",
        "deploy",
        "admin",
        "teardown",
        "docker",
        "initializer",
        "sidecar",
        "proxy",
        "notify",
        "upgrade",
        "port-forward",
        "sandbox",
    ]
    if not (
        context.invoked_subcommand in non_check_cmds
        or offline
        or settings.CLIENT_CONFIG.no_api
        or PolyaxonServices.get_service_name()
        or DOCS_GEN
    ) and not (settings.CLI_CONFIG or settings.CLI_CONFIG.installation):
        cli_config = set_versions_config(is_cli=False)
        settings.CLI_CONFIG = cli_config
        check_cli_version(cli_config, is_cli=False)


cli.add_command(login)
cli.add_command(logout)
cli.add_command(whoami)
cli.add_command(upgrade)
cli.add_command(version)
cli.add_command(config)
cli.add_command(check)
cli.add_command(init)
cli.add_command(project)
cli.add_command(ops)
cli.add_command(artifacts)
cli.add_command(components)
cli.add_command(models)
cli.add_command(run)
cli.add_command(dashboard)
cli.add_command(admin)
cli.add_command(port_forward)
cli.add_command(completion)

# INIT
if PolyaxonServices.is_init():
    from polyaxon._cli.services.clean_artifacts import clean_artifacts
    from polyaxon._cli.services.docker import docker
    from polyaxon._cli.services.initializer import initializer
    from polyaxon._cli.services.wait import wait

    cli.add_command(clean_artifacts)
    cli.add_command(docker)
    cli.add_command(initializer)
    cli.add_command(wait)

# Events
if PolyaxonServices.is_events_handlers():
    from polyaxon._cli.services.notifier import notify

    cli.add_command(notify)

# Sidecar
if PolyaxonServices.is_sidecar():
    from polyaxon._cli.services.sidecar import sidecar

    cli.add_command(sidecar)

# Tuner
if PolyaxonServices.is_hp_search():
    from polyaxon._cli.services.tuner import tuner

    cli.add_command(tuner)

# Agents
if PolyaxonServices.is_agent():
    from polyaxon._cli.services.agent import agent

    cli.add_command(agent)

try:
    from haupt.cli.sandbox import sandbox

    cli.add_command(sandbox)
except ImportError:
    pass



from typing import Optional

MAIN_JOB_CONTAINER = "polyaxon-main"
SIDECAR_CONTAINER = "polyaxon-sidecar"
TFJOBS_CONTAINER = "tensorflow"
PYTORCHJOBS_CONTAINER = "pytorch"
MXJOBS_CONTAINER = "mxnet"
XGBJOBS_CONTAINER = "xgboost"
PADDLEJOBS_CONTAINER = "paddle"
MAIN_CONTAINER_NAMES = [
    MAIN_JOB_CONTAINER,
    TFJOBS_CONTAINER,
    PYTORCHJOBS_CONTAINER,
    MXJOBS_CONTAINER,
    XGBJOBS_CONTAINER,
    PADDLEJOBS_CONTAINER,
]
INIT_AUTH_CONTAINER = "polyaxon-init-auth"
INIT_DOCKERFILE_CONTAINER_PREFIX = "polyaxon-init-dockerfile"
INIT_FILE_CONTAINER_PREFIX = "polyaxon-init-file"
INIT_TENSORBOARD_CONTAINER_PREFIX = "polyaxon-init-tb"
INIT_GIT_CONTAINER_PREFIX = "polyaxon-init-git"
INIT_CUSTOM_CONTAINER_PREFIX = "polyaxon-init-custom"
INIT_ARTIFACTS_CONTAINER_PREFIX = "polyaxon-init-artifacts"
POLYAXON_INIT_PREFIX = "polyaxon-init"
INIT_PREFIX = "init"
SIDECAR_PREFIX = "sidecar"


def generate_container_name(
    prefix: str, suffix: Optional[str] = None, unique: bool = True
) -> str:
    import uuid

    prefix = prefix or "container"
    unique_value = uuid.uuid4().hex[:10]
    if suffix:
        suffix = suffix.replace("_", "-")
        if unique:
            suffix = "{}-{}".format(suffix, unique_value)
    else:
        suffix = unique_value
    return "{}-{}".format(prefix, suffix)


def sanitize_container_name(name: str) -> str:
    name = name.replace("_", "-")
    return name.lower()



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from clipped.compact.pydantic import Field, PositiveInt, validator
from clipped.config.schema import skip_partial
from clipped.types.ref_or_obj import RefField

from polyaxon._flow.early_stopping import V1EarlyStopping
from polyaxon._flow.matrix.base import BaseSearchConfig
from polyaxon._flow.matrix.enums import V1MatrixKind
from polyaxon._flow.matrix.params import V1HpParam


def validate_matrix(matrix):
    if not matrix:
        return None

    for key, value in matrix.items():
        if value.is_distribution:
            raise ValueError(
                "`{}` defines a distribution, "
                "and it cannot be used with grid search.".format(key)
            )

    return matrix


class V1GridSearch(BaseSearchConfig):
    """Grid search is essentially an exhaustive search through a manually
    specified set of hyperparameters.

    User can possibly limit the number of experiments and not traverse the whole
    search space created by providing `numRuns`.

    Grid search does not allow the use of distributions,
    and requires that all values of the params definition to
    be [discrete values](/docs/automation/optimization-engine/params/#discrete-values).

    Args:
        kind: str, should be equal `grid`
        params: List[Dict[str, [params](/docs/automation/optimization-engine/params/#discrete-values)]]  # noqa
        concurrency: int, optional
        num_runs: int, optional
        early_stopping: List[[EarlyStopping](/docs/automation/helpers/early-stopping)], optional

    ## YAML usage

    ```yaml
    >>> matrix:
    >>>   kind: grid
    >>>   concurrency:
    >>>   params:
    >>>   numRuns:
    >>>   earlyStopping:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import (
    >>>     V1GridSearch, V1HpLogSpace, V1HpChoice, V1FailureEarlyStopping, V1MetricEarlyStopping
    >>> )
    >>> matrix = V1GridSearch(
    >>>   concurrency=2,
    >>>   params={"param1": V1HpLogSpace(...), "param2": V1HpChoice(...), ... },
    >>>   num_runs=5
    >>>   early_stopping=[V1FailureEarlyStopping(...), V1MetricEarlyStopping(...)]
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this matrix is a grid search.

    If you are using the python client to create the mapping,
    this field is not required and is set by default.

    ```yaml
    >>> matrix:
    >>>   kind: grid
    ```

    ### concurrency

    An optional value to set the number of concurrent operations.

    <blockquote class="light">
    This value only makes sense if less or equal to the total number of possible runs.
    </blockquote>

    ```yaml
    >>> matrix:
    >>>   kind: grid
    >>>   concurrency: 2
    ```

    For more details about concurrency management,
    please check the [concurrency section](/docs/automation/helpers/concurrency/).

    ### params

    A dictionary of `key -> value generator`
    to generate the parameters.

    Gird search can only use
    [discrete value](/docs/automation/optimization-engine/params/#discrete-values).

    > The parameters generated will be validated against
    > the component's inputs/outputs definition to check that the values
    > can be passed and have valid types.

    ```yaml
    >>> matrix:
    >>>   kind: grid
    >>>   params:
    >>>     param1:
    >>>        kind: ...
    >>>        value: ...
    >>>     param2:
    >>>        kind: ...
    >>>        value: ...
    ```


    ### numRuns

    Maximum number of runs to start based on the search space defined.

    ```yaml
    >>> matrix:
    >>>   kind: grid
    >>>   numRuns: 5
    ```

    ### earlyStopping

    A list of early stopping conditions to check for terminating
    all operations managed by the pipeline.
    If one of the early stopping conditions is met,
    a signal will be sent to terminate all running and pending operations.

    ```yaml
    >>> matrix:
    >>>   kind: grid
    >>>   earlyStopping: ...
    ```

    For more details please check the
    [early stopping section](/docs/automation/helpers/early-stopping/).

    ## Example

    This example will define 10 experiments based on the cartesian product
    of `lr` and `dropout` possible values.

    ```yaml
    >>> version: 1.1
    >>> kind: operation
    >>> matrix:
    >>>   kind: grid
    >>>   concurrency: 2
    >>>   params:
    >>>     lr:
    >>>       kind: logspace
    >>>       value: 0.01:0.1:5
    >>>     dropout:
    >>>       kind: choice
    >>>       value: [0.2, 0.5]
    >>>    early_stopping:
    >>>      - metric: accuracy
    >>>        value: 0.9
    >>>        optimization: maximize
    >>>      - metric: loss
    >>>        value: 0.05
    >>>        optimization: minimize
    >>> component:
    >>>   inputs:
    >>>     - name: batch_size
    >>>       type: int
    >>>       isOptional: true
    >>>       value: 128
    >>>     - name: lr
    >>>       type: float
    >>>     - name: dropout
    >>>       type: float
    >>>   container:
    >>>     image: image:latest
    >>>     command: [python3, train.py]
    >>>     args: ["--batch-size={{ batch_size }}", "--lr={{ lr }}", "--dropout={{ dropout }}"]
    ```
    """

    _IDENTIFIER = V1MatrixKind.GRID

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    params: Union[Dict[str, V1HpParam], RefField]
    num_runs: Optional[Union[PositiveInt, RefField]] = Field(alias="numRuns")
    concurrency: Optional[Union[PositiveInt, RefField]]
    early_stopping: Optional[Union[List[V1EarlyStopping], RefField]] = Field(
        alias="earlyStopping"
    )

    @validator("num_runs", "concurrency", pre=True)
    def check_values(cls, v, field):
        if v and v < 1:
            raise ValueError(f"{field} must be greater than 1, received `{v}` instead.")
        return v

    @validator("params", always=True)
    @skip_partial
    def validate_matrix(cls, params):
        return validate_matrix(params)



import jinja2

from typing import Optional

from markupsafe import soft_str


def map_format(value, pattern, variable_name: Optional[str] = None):
    if variable_name:
        return soft_str(pattern) % {variable_name: value}
    return soft_str(pattern) % value


def get_engine():
    env = jinja2.Environment()
    env.filters["map_format"] = map_format
    return env



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr


class V1EntitiesTransfer(BaseAllowSchemaModel):
    uuids: Optional[List[UUIDStr]]
    project: Optional[StrictStr]



from typing import Optional

from urllib3.exceptions import HTTPError

from polyaxon._client.client import PolyaxonClient
from polyaxon._contexts import paths as ctx_paths
from polyaxon._schemas.authentication import AccessTokenConfig
from polyaxon.exceptions import ApiException, PolyaxonClientException


def create_context_auth(
    access_token: AccessTokenConfig, context_auth_path: Optional[str] = None
):
    context_auth_path = context_auth_path or ctx_paths.CONTEXT_MOUNT_AUTH
    with open(context_auth_path, "w") as config_file:
        config_file.write(access_token.to_json())


def impersonate(
    owner: str, project: str, run_uuid: str, client: Optional[PolyaxonClient] = None
):
    try:
        client = client or PolyaxonClient()
        response = client.runs_v1.impersonate_token(owner, project, run_uuid)
        polyaxon_client = PolyaxonClient(token=response.token)
        user = polyaxon_client.users_v1.get_user()
        access_token = AccessTokenConfig.construct(
            username=user.username, token=response.token
        )
        create_context_auth(access_token)
    except (ApiException, HTTPError) as e:
        raise PolyaxonClientException(
            "This worker is not allowed to run this job %s." % e
        )



import sys

from typing import Tuple

from clipped.formatting import Printer

from polyaxon._constants.globals import DEFAULT, DEFAULT_HUB
from polyaxon._env_vars.getters.user import get_local_owner
from polyaxon.exceptions import PolyaxonClientException, PolyaxonSchemaError


def get_versioned_entity_info(
    entity: str, entity_name: str, default_owner: str
) -> Tuple[str, str, str]:
    if not entity:
        raise PolyaxonSchemaError(
            "Received an invalid {} reference: `{}`".format(entity_name, entity)
        )
    entity_values = entity.split(":")
    if len(entity_values) > 2:
        raise PolyaxonSchemaError(
            "Received an invalid {} reference: `{}`".format(entity_name, entity)
        )
    if len(entity_values) == 2:
        entity_namespace, version = entity_values
    else:
        entity_namespace, version = entity_values[0], "latest"
    version = version or "latest"
    parts = entity_namespace.replace(".", "/").split("/")
    owner = default_owner
    if not parts or len(parts) > 2:
        raise PolyaxonSchemaError(
            "Received an invalid {} reference: `{}`".format(entity_name, entity)
        )
    if len(parts) == 2:
        owner, entity_namespace = parts
    return owner, entity_namespace, version


def get_component_info(hub: str, use_local_owner: bool = False) -> Tuple[str, str, str]:
    from polyaxon import settings

    if use_local_owner:
        try:
            owner = get_local_owner()
        except PolyaxonClientException:
            owner = None

        if not owner and (not settings.CLI_CONFIG or settings.CLI_CONFIG.is_community):
            owner = DEFAULT
    else:
        owner = DEFAULT_HUB
    return get_versioned_entity_info(
        entity=hub, entity_name="component", default_owner=owner
    )


def get_model_info(entity: str, entity_name: str, is_cli: bool = False):
    from polyaxon import settings

    if not entity:
        message = "Please provide a valid {}!".format(entity_name)
        if is_cli:
            Printer.error(message)
            sys.exit(1)
        else:
            raise PolyaxonClientException(message)

    try:
        owner = get_local_owner()
    except PolyaxonClientException:
        owner = None

    if not owner and (not settings.CLI_CONFIG or settings.CLI_CONFIG.is_community):
        owner = DEFAULT

    owner, entity_namespace, version = get_versioned_entity_info(
        entity=entity, entity_name=entity_name, default_owner=owner
    )

    if not owner:
        owner = settings.AUTH_CONFIG.username if settings.AUTH_CONFIG else None

    if not all([owner, entity_name]):
        message = "Please provide a valid {}.".format(entity_name)
        if is_cli:
            Printer.error(message)
            sys.exit(1)
        else:
            raise PolyaxonClientException(message)
    return owner, entity_namespace, version



from typing import Any, Dict, List, Optional

from clipped.config.schema import BaseAllowSchemaModel


class V1EventsResponse(BaseAllowSchemaModel):
    data: Optional[List[Dict[str, Any]]]


class V1MultiEventsResponse(BaseAllowSchemaModel):
    data: Optional[Dict[str, Any]]



from typing import Optional

from polyaxon import pkg
from polyaxon._containers.names import MAIN_JOB_CONTAINER
from polyaxon._containers.pull_policy import PullPolicy
from polyaxon._k8s import k8s_schemas
from polyaxon._k8s.k8s_schemas import V1Container


def get_default_tuner_container(
    command, bracket_iteration: Optional[int] = None
) -> V1Container:
    args = [
        "{{params.matrix.as_arg}}",
        "{{params.join.as_arg}}",
        "{{params.iteration.as_arg}}",
    ]
    if bracket_iteration is not None:
        args.append("{{params.bracket_iteration.as_arg}}")
    return V1Container(
        name=MAIN_JOB_CONTAINER,
        image="polyaxon/polyaxon-hpsearch:{}".format(pkg.VERSION),
        image_pull_policy=PullPolicy.IF_NOT_PRESENT.value,
        command=command,
        args=args,
        resources=k8s_schemas.V1ResourceRequirements(
            requests={"cpu": "0.1", "memory": "180Mi"},
        ),
    )



K8S_API_VERSION_V1 = "v1"
K8S_API_VERSION_NETWORKING_V1_BETA1 = "networking.k8s.io/v1beta1"
K8S_API_VERSION_BATCH_V1 = "batch/v1"
K8S_API_VERSION_APPS_V1 = "apps/v1"
K8S_PERSISTENT_VOLUME_KIND = "PersistentVolume"
K8S_PERSISTENT_VOLUME_CLAIM_KIND = "PersistentVolumeClaim"
K8S_CONFIG_MAP_KIND = "ConfigMap"
K8S_SECRET_KIND = "Secret"  # noqa
K8S_POD_KIND = "Pod"
K8S_POD_TEMPLATE_KIND = "PodTemplate"
K8S_DEPLOYMENT_KIND = "Deployment"
K8S_SERVICE_KIND = "Service"
K8S_INGRESS_KIND = "Ingress"
K8S_JOB_KIND = "Job"



from polyaxon._deploy.operators.cmd_operator import CmdOperator
from polyaxon.exceptions import PolyaxonOperatorException
from polyaxon.logger import logger


class DockerOperator(CmdOperator):
    CMD = "docker"

    @classmethod
    def params(cls, args):
        params = [cls.CMD] + args
        return params

    @classmethod
    def check(cls):
        command_exist = cls.execute(args=["version"])
        if not command_exist:
            return False
        return True

    @classmethod
    def execute(cls, args, env=None, stream=False, output_only: bool = True):
        params = cls.params(args)
        logger.debug("Docker operator executing command: {}".format(" ".join(args)))
        return cls._execute(
            params=params, env=env, stream=stream, output_only=output_only
        )

    @classmethod
    def set_volume(cls, volume):
        args = ["volume", "create", "--name={}".format(volume)]
        if not volume:
            raise PolyaxonOperatorException(
                "docker", args, None, None, "Volume name was not provided"
            )
        return cls.execute(args)



from typing import List, Optional, Union

from clipped.compact.pydantic import Field, StrictStr, constr, validator
from clipped.types.ref_or_obj import BoolOrRef, FloatOrRef, RefField
from clipped.utils.lists import to_list

from polyaxon._flow.builds import V1Build
from polyaxon._flow.cache import V1Cache
from polyaxon._flow.hooks import V1Hook
from polyaxon._flow.plugins import V1Plugins
from polyaxon._flow.termination import V1Termination
from polyaxon._schemas.base import NAME_REGEX, BaseSchemaModel


class BaseComponent(BaseSchemaModel):
    version: Optional[float]
    kind: Optional[StrictStr]
    name: Optional[Union[constr(regex=NAME_REGEX), RefField]]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    presets: Optional[List[StrictStr]]
    queue: Optional[StrictStr]
    namespace: Optional[StrictStr]
    cache: Optional[Union[V1Cache, RefField]]
    termination: Optional[Union[V1Termination, RefField]]
    plugins: Optional[Union[V1Plugins, RefField]]
    build: Optional[Union[V1Build, RefField]]
    hooks: Optional[Union[List[V1Hook], RefField]]
    is_approved: Optional[BoolOrRef] = Field(alias="isApproved")
    cost: Optional[FloatOrRef]

    @validator("tags", "presets", pre=True)
    def validate_str_list(cls, v):
        if isinstance(v, str):
            return to_list(v, check_str=True)
        return v



from typing import List, Optional, Union

from clipped.compact.pydantic import StrictStr
from clipped.types.ref_or_obj import RefField

from polyaxon._schemas.types.base import BaseTypeConfig


class V1GitType(BaseTypeConfig):
    """Git type allows you to pass a git repo as a parameter.

    If used as an input type, Polyaxon can resolve several git connections
    and will turn this input type into an initializer with logic to clone
    the repo with support for branches and commits,
    the requested repo will be exposed as a context for your jobs and operations.

    Args:
        url: str, optional.
        revision: str, optional.
        flags: List[str], optional

    ### YAML usage

    ### Usage in IO and params definition

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: git
    >>>   - name: test2
    >>>     type: git
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: {"url": "https://github.com/tensorflow/models"}}
    >>>   test2: {value: {revision: "branchA"}, connection: "my-git-connection"}
    >>>   test3: {
    >>>     value: {flags: ["--recursive", "-c http.sslVerify=false"]},
    >>>     connection: "my-git-connection",
    >>>   }
    ```

    ### Usage in initializers

    ```yaml
    >>> version:  1.1
    >>> kind: component
    >>> run:
    >>>   kind: job
    >>>   init:
    >>>   - git: {"url": "https://github.com/tensorflow/models"}
    >>>   - git:
    >>>       revision: branchA
    >>>     connection: my-git-connection
    >>>   - git:
    >>>       flags: ["--recursive", "-c http.sslVerify=false"]
    >>>     connection: my-git-connection
    >>>   ...
    ```

    ### Python usage

    ### Usage in IO and params definition

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.GIT,
    >>>     ),
    >>>     V1IO(
    >>>         name="test2",
    >>>         type=types.GIT,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1Param
    >>> params = {
    >>>     "test1": V1Param(
    >>>         value=types.V1GitType(url="https://github.com/tensorflow/models")
    >>>     ),
    >>>     "test2": V1Param(
    >>>         value=types.V1GitType(revision="branchA"),
    >>>         connection="my-git-connection",
    >>>     ),
    >>> }
    ```

    ### Usage in initializers
    ```python
    >>> from polyaxon.schemas import V1Component, V1Init, V1Job
    >>> from polyaxon.types import V1GitType
    >>> from polyaxon import k8s
    >>> component = V1Component(
    >>>     run=V1Job(
    >>>        init=[
    >>>             V1Init(
    >>>               git=V1GitType(url="https://github.com/tensorflow/models"),
    >>>             ),
    >>>             V1Init(
    >>>               git=V1GitType(revision="branchA"),
    >>>               connection="my-git-connection",
    >>>             ),
    >>>        ],
    >>>        container=k8s.V1Container(...)
    >>>     )
    >>> )
    ```
    """

    _IDENTIFIER = "git"

    url: Optional[StrictStr]
    revision: Optional[StrictStr]
    flags: Optional[Union[List[StrictStr], RefField]]

    def get_name(self):
        if self.url:
            return self.url.split("/")[-1].split(".")[0]
        return None



from typing import Dict, List

from clipped.utils.lists import to_list

from polyaxon._connections import V1Connection
from polyaxon._flow import V1CompiledOperation, V1Init


class BaseContextsManager:
    @classmethod
    def _resolve_init_contexts(
        cls,
        contexts: Dict,
        init: List[V1Init],
        connection_by_names: Dict[str, V1Connection],
    ):
        init = to_list(init, check_none=True)
        connections = [i.connection for i in init if i.connection]
        return cls._resolve_connections_contexts(
            contexts=contexts,
            connections=connections,
            connection_by_names=connection_by_names,
            key="init",
        )

    @staticmethod
    def _resolve_connections_contexts(
        contexts: Dict,
        connections: List[str],
        connection_by_names: Dict[str, V1Connection],
        key: str = "connections",
    ) -> Dict:
        connections = to_list(connections, check_none=True)
        for connection in connections:
            if connection_by_names[connection].schema_:
                schema_ = connection_by_names[connection].schema_
                if hasattr(schema_, "to_dict"):
                    contexts[key][connection] = schema_.to_dict()
                else:
                    contexts[key][connection] = schema_
            else:
                contexts[key][connection] = {}
        return contexts

    @classmethod
    def _resolver_replica(
        cls,
        contexts: Dict,
        init: List[V1Init],
        connections: List[str],
        connection_by_names: Dict[str, V1Connection],
    ) -> Dict:
        contexts["init"] = {}
        contexts["connections"] = {}
        contexts = cls._resolve_init_contexts(
            contexts=contexts, init=init, connection_by_names=connection_by_names
        )
        contexts = cls._resolve_connections_contexts(
            contexts=contexts,
            connections=connections,
            connection_by_names=connection_by_names,
        )
        return contexts

    @classmethod
    def resolve(
        cls,
        namespace: str,
        owner_name: str,
        project_name: str,
        run_uuid: str,
        contexts: Dict,
        compiled_operation: V1CompiledOperation,
        connection_by_names: Dict[str, V1Connection],
    ) -> Dict:
        raise NotImplementedError()



from polyaxon._client.client import PolyaxonClient
from polyaxon._client.decorators import ensure_is_managed
from polyaxon._client.project import ProjectClient
from polyaxon._client.run import RunClient, get_run_logs
from polyaxon._client.store import PolyaxonStore
from polyaxon._schemas.agent import AgentConfig
from polyaxon._schemas.authentication import AccessTokenConfig
from polyaxon._schemas.cli import CliConfig
from polyaxon._schemas.client import ClientConfig
from polyaxon._sdk.api import (
    AgentsV1Api,
    ArtifactsStoresV1Api,
    AuthV1Api,
    ConnectionsV1Api,
    DashboardsV1Api,
    OrganizationsV1Api,
    PresetsV1Api,
    ProjectDashboardsV1Api,
    ProjectSearchesV1Api,
    ProjectsV1Api,
    QueuesV1Api,
    RunsV1Api,
    SearchesV1Api,
    ServiceAccountsV1Api,
    TagsV1Api,
    TeamsV1Api,
    UsersV1Api,
    VersionsV1Api,
)
from polyaxon._sdk.async_client.api_client import AsyncApiClient
from polyaxon._sdk.configuration import Configuration
from polyaxon._sdk.sync_client.api_client import ApiClient
from polyaxon.schemas import *



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel

from traceml.artifacts import V1RunArtifact


class V1ListRunArtifactsResponse(BaseAllowSchemaModel):
    count: Optional[int]
    results: Optional[List[V1RunArtifact]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



import sys

import click

from clipped.formatting import Printer
from clipped.utils.dicts import dict_to_tabulate
from urllib3.exceptions import HTTPError

from polyaxon import settings
from polyaxon._cli.dashboard import get_dashboard_url
from polyaxon._cli.errors import handle_cli_error, handle_command_not_in_ce, not_in_ce
from polyaxon._cli.session import (
    ensure_cli_config,
    session_expired,
    set_versions_config,
)
from polyaxon._managers.auth import AuthConfigManager
from polyaxon._managers.cli import CliConfigManager
from polyaxon._managers.user import UserConfigManager
from polyaxon._schemas.authentication import AccessTokenConfig, V1Credentials
from polyaxon._sdk.schemas.v1_auth import V1Auth
from polyaxon._sdk.schemas.v1_user import V1User
from polyaxon.client import PolyaxonClient
from polyaxon.exceptions import ApiException
from polyaxon.logger import clean_outputs, logger


def get_user_info(user: V1User):
    response = dict_to_tabulate(user.to_dict(), exclude_attrs=["role", "theme"])
    Printer.heading("User info:")
    Printer.dict_tabulate(response)


@click.command()
@click.option("--token", "-t", help="Polyaxon token.")
@click.option("--username", "-u", help="Polyaxon username or email.")
@click.option("--password", "-p", help="Polyaxon password.")
@clean_outputs
def login(token, username, password):
    """Login to Polyaxon Cloud or Polyaxon EE."""
    polyaxon_client = PolyaxonClient()
    ensure_cli_config()

    if not settings.CLI_CONFIG or settings.CLI_CONFIG.is_community:
        handle_command_not_in_ce()

    if username and not token:
        # Use user or email / password login
        if not password:
            password = click.prompt(
                "Please enter your password", type=str, hide_input=True
            )
            password = password.strip()
            if not password:
                logger.info(
                    "You entered an empty string. "
                    "Please make sure you enter your password correctly."
                )
                sys.exit(1)

        try:
            body = V1Credentials(username=username, password=password)
            access_auth = polyaxon_client.auth_v1.login(body=body)
        except (ApiException, HTTPError) as e:
            AuthConfigManager.purge()
            UserConfigManager.purge()
            CliConfigManager.purge()
            handle_cli_error(e, message="Could not login.")
            sys.exit(1)

        if not access_auth.token:
            Printer.error("Failed to login")
            return
    else:
        if not token:
            token_url = get_dashboard_url(subpath="profile/token")
            click.confirm(
                "Authentication token page will now open in your browser. Continue?",
                abort=True,
                default=True,
            )

            click.launch(token_url)
            logger.info("Please copy and paste the authentication token.")
            token = click.prompt(
                "This is an invisible field. Paste token and press ENTER",
                type=str,
                hide_input=True,
            )

        if not token:
            logger.info(
                "Empty token received. "
                "Make sure your shell is handling the token appropriately."
            )
            logger.info(
                "See docs for help: http://polyaxon.com/docs/polyaxon_cli/commands/auth"
            )
            return

        access_auth = V1Auth(token=token.strip(" "))

    # Set user
    try:
        AuthConfigManager.purge()
        UserConfigManager.purge()
        polyaxon_client = PolyaxonClient(token=access_auth.token)
        user = polyaxon_client.users_v1.get_user()
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not load user info.")
        sys.exit(1)
    access_token = AccessTokenConfig(username=user.username, token=access_auth.token)
    AuthConfigManager.set_config(access_token)
    UserConfigManager.set_config(user)
    polyaxon_client.config.token = access_auth.token
    Printer.success("Login successful")
    get_user_info(user)

    set_versions_config(polyaxon_client=polyaxon_client, set_handler=True)


@click.command()
@clean_outputs
def logout():
    """Logout from Polyaxon Cloud or Polyaxon EE."""
    AuthConfigManager.purge()
    UserConfigManager.purge()
    CliConfigManager.purge()
    Printer.success("You are logged out")


@click.command()
@not_in_ce
@clean_outputs
def whoami():
    """Show current logged Polyaxon Cloud or Polyaxon EE user."""
    try:
        polyaxon_client = PolyaxonClient()
        user = polyaxon_client.users_v1.get_user()
        get_user_info(user)
    except ApiException as e:
        if e.status == 403:
            session_expired()
        handle_cli_error(e, message="Could not get the user info.", sys_exit=True)
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not load user info.", sys_exit=True)



from clipped.formatting import Printer
from clipped.utils.dicts import dict_to_tabulate
from clipped.utils.json import orjson_dumps


def get_entity_details(entity: any, entity_name: str):
    if entity.description:
        Printer.heading("{} description:".format(entity_name))
        Printer.print("{}\n".format(entity.description))

    if entity.settings:
        Printer.heading("{} settings:".format(entity_name))
        Printer.print(
            "{}\n".format(
                entity.settings.to_dict()
                if hasattr(entity.settings, "to_dict")
                else entity.settings
            )
        )

    if entity.readme:
        Printer.heading("{} readme:".format(entity_name))
        Printer.print_md(entity.readme)

    response = dict_to_tabulate(
        entity.to_dict(),
        humanize_values=True,
        exclude_attrs=["description", "settings", "readme"],
    )

    Printer.heading("{} info:".format(entity_name))
    Printer.dict_tabulate(response)


def handle_output(response: any, output: str):
    if output == "json":
        Printer.pprint(response)
        return
    if "path=" in output:
        json_path = output.strip("path=")
        with open(json_path, "w", encoding="utf8", newline="") as output_file:
            output_file.write(orjson_dumps(response))



import os
import subprocess
import sys

import click

from clipped.compact.pydantic import ValidationError
from clipped.formatting import Printer
from clipped.utils.lists import to_list
from clipped.utils.paths import check_or_create_path, copy_file

from polyaxon._config.spec import ConfigSpec
from polyaxon._connections import V1ConnectionKind
from polyaxon._schemas.types import V1FileType
from polyaxon.exceptions import PolyaxonSchemaError


@click.group()
def initializer():
    pass


@initializer.command()
def auth():
    """Create auth context."""
    from polyaxon._init.auth import create_auth_context

    create_auth_context()


@initializer.command()
@click.option("--file-context", help="The context file definition.")
@click.option("--filepath", help="The path where to save the file.")
@click.option("--copy-path", help="Copy generated files to a specific path.")
@click.option(
    "--track",
    is_flag=True,
    default=False,
    help="Flag to track or not the file content.",
)
def file(file_context, filepath, copy_path, track):
    """Create auth context."""
    from clipped.utils.hashing import hash_value

    from polyaxon._init.file import create_file_lineage

    try:
        file_context = V1FileType.from_dict(ConfigSpec.read_from(file_context))
    except (PolyaxonSchemaError, ValidationError) as e:
        Printer.error("received a non valid file context.")
        Printer.error("Error message: {}.".format(e))
        sys.exit(1)

    filepath = os.path.join(filepath, file_context.filename)
    check_or_create_path(filepath, is_dir=False)
    # Clean any previous file on that path
    if os.path.exists(filepath):
        os.remove(filepath)

    with open(filepath, "w") as generated_file:
        generated_file.write(file_context.content)
        if file_context.chmod:
            subprocess.check_call(["chmod", file_context.chmod, filepath])

    if copy_path:
        filepath = copy_file(filepath, copy_path)

    if track:
        create_file_lineage(
            filepath=filepath,
            summary={"hash": hash_value(file_context.content)},
            kind=file_context.kind,
        )

    Printer.success("File is initialized, path: `{}`".format(filepath))


@initializer.command()
@click.option("--url", help="The git url to pull.")
@click.option("--revision", help="The revision(commit/branch/treeish) to pull.")
@click.option("--repo-path", help="The path to where to pull the repos.")
@click.option("--connection", help="The connection used for pulling this repo.")
@click.option("--flags", help="Additional flags for pulling this repo.")
def git(url, repo_path, revision, connection, flags):
    """Create auth context."""
    from polyaxon._init.git import create_code_repo

    if flags:
        flags = to_list(flags, check_str=True)

    create_code_repo(
        repo_path=repo_path,
        url=url,
        revision=revision,
        connection=connection,
        flags=flags,
    )

    Printer.success("Git Repo is initialized, path: `{}`".format(repo_path))


@initializer.command()
@click.option("--connection-name", help="The connection name.")
@click.option("--path-from", help="The s3 path to download data from.")
@click.option("--path-to", help="The local path to store the data.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
@click.option(
    "--check-path",
    is_flag=True,
    default=False,
    help="whether check the path if file or dir.",
)
@click.option(
    "--raise-errors",
    is_flag=True,
    default=False,
    help="whether or not to raise initialization errors.",
)
@click.option(
    "--sync-fw",
    is_flag=True,
    default=False,
    help="whether or not to sync file watcher after initialization.",
)
def s3(connection_name, path_from, path_to, is_file, check_path, raise_errors, sync_fw):
    """Create s3 path context."""
    from polyaxon._init.artifacts import download_artifact

    download_artifact(
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.S3,
        path_from=path_from,
        path_to=path_to,
        is_file=is_file,
        raise_errors=raise_errors,
        sync_fw=sync_fw,
        check_path=check_path,
    )


@initializer.command()
@click.option("--connection-name", help="The connection name.")
@click.option("--path-from", help="The gcs path to download data from.")
@click.option("--path-to", help="The local path to store the data.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
@click.option(
    "--check-path",
    is_flag=True,
    default=False,
    help="whether check the path if file or dir.",
)
@click.option(
    "--raise-errors",
    is_flag=True,
    default=False,
    help="whether or not to raise initialization errors.",
)
@click.option(
    "--sync-fw",
    is_flag=True,
    default=False,
    help="whether or not to sync file watcher after initialization.",
)
def gcs(
    connection_name, path_from, path_to, is_file, check_path, raise_errors, sync_fw
):
    """Create gcs path context."""
    from polyaxon._init.artifacts import download_artifact

    download_artifact(
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.GCS,
        path_from=path_from,
        path_to=path_to,
        is_file=is_file,
        raise_errors=raise_errors,
        sync_fw=sync_fw,
        check_path=check_path,
    )


@initializer.command()
@click.option("--connection-name", help="The connection name.")
@click.option("--path-from", help="The wasb path to download data from.")
@click.option("--path-to", help="The local path to store the data.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
@click.option(
    "--check-path",
    is_flag=True,
    default=False,
    help="whether check the path if file or dir.",
)
@click.option(
    "--raise-errors",
    is_flag=True,
    default=False,
    help="whether or not to raise initialization errors.",
)
@click.option(
    "--sync-fw",
    is_flag=True,
    default=False,
    help="whether or not to sync file watcher after initialization.",
)
def wasb(
    connection_name, path_from, path_to, is_file, check_path, raise_errors, sync_fw
):
    """Create wasb path context."""
    from polyaxon._init.artifacts import download_artifact

    download_artifact(
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.WASB,
        path_from=path_from,
        path_to=path_to,
        is_file=is_file,
        raise_errors=raise_errors,
        sync_fw=sync_fw,
        check_path=check_path,
    )


@initializer.command()
@click.option("--connection-kind", help="The connection kind.")
@click.option("--connection-name", help="The connection name.")
@click.option("--path-from", help="The path to download data from.")
@click.option("--path-to", help="The local path to store the data.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
@click.option(
    "--check-path",
    is_flag=True,
    default=False,
    help="whether check the path if file or dir.",
)
@click.option(
    "--raise-errors",
    is_flag=True,
    default=False,
    help="whether or not to raise initialization errors.",
)
@click.option(
    "--sync-fw",
    is_flag=True,
    default=False,
    help="whether or not to sync file watcher after initialization.",
)
def path(
    connection_kind,
    connection_name,
    path_from,
    path_to,
    is_file,
    check_path,
    raise_errors,
    sync_fw,
):
    """Create path context."""
    from polyaxon._init.artifacts import download_artifact

    download_artifact(
        connection_name=connection_name,
        connection_kind=connection_kind,
        path_from=path_from,
        path_to=path_to,
        is_file=is_file,
        raise_errors=raise_errors,
        sync_fw=sync_fw,
        check_path=check_path,
    )


@initializer.command()
@click.option("--port", type=int, help="The connection kind.")
@click.option("--connection-kind", help="The connection kind.")
@click.option("--connection-name", help="The connection name.")
@click.option(
    "--context-from", help="The context path where to load the tensorboard logs from."
)
@click.option("--context-to", help="The context path to store the tensorboard logs.")
@click.option(
    "--uuids",
    "--uuid",
    "-uid",
    help="The operation uuids to initialize.",
)
@click.option(
    "--use-names", is_flag=True, help="Use run names to initialize hte paths."
)
@click.option("--path-prefix", help="The operation name to initialize.")
@click.option("--plugins", help="The operation uuids to initialize.")
def tensorboard(
    port,
    connection_kind,
    connection_name,
    context_from,
    context_to,
    uuids,
    use_names,
    path_prefix,
    plugins,
):
    """Create path context."""
    from clipped.utils.validation import validate_tags

    from polyaxon._init.tensorboard import initialize_tensorboard

    uuids = validate_tags(uuids)
    plugins = validate_tags(plugins)
    initialize_tensorboard(
        port=port,
        connection_name=connection_name,
        connection_kind=connection_kind,
        context_from=context_from,
        context_to=context_to or ".",
        uuids=uuids,
        use_names=use_names,
        path_prefix=path_prefix,
        plugins=plugins,
    )


@initializer.command()
@click.option("--path", help="The local path to store the data.")
def fswatch(path):
    from polyaxon._init.artifacts import sync_file_watcher

    sync_file_watcher(path)



from clipped.utils.enums import PEnum


class V1MatrixKind(str, PEnum):
    RANDOM = "random"
    GRID = "grid"
    HYPERBAND = "hyperband"
    BAYES = "bayes"
    HYPEROPT = "hyperopt"
    ITERATIVE = "iterative"
    MAPPING = "mapping"

    @classmethod
    def iteration_values(cls):
        return {
            cls.HYPERBAND,
            cls.BAYES,
            cls.HYPEROPT,
            cls.ITERATIVE,
        }


class V1HPKind(str, PEnum):
    CHOICE = "choice"
    PCHOICE = "pchoice"
    RANGE = "range"
    DATERANGE = "daterange"
    DATETIMERANGE = "datetimerange"
    LINSPACE = "linspace"
    LOGSPACE = "logspace"
    GEOMSPACE = "geomspace"
    UNIFORM = "uniform"
    QUNIFORM = "quniform"
    LOGUNIFORM = "loguniform"
    QLOGUNIFORM = "qloguniform"
    NORMAL = "normal"
    QNORMAL = "qnormal"
    LOGNORMAL = "lognormal"
    QLOGNORMAL = "qlognormal"


class AcquisitionFunctions(str, PEnum):
    UCB = "ucb"
    EI = "ei"
    POI = "poi"

    @classmethod
    def ucb_values(cls):
        return {cls.UCB, cls.UCB.upper(), cls.UCB.capitalize()}

    @classmethod
    def ei_values(cls):
        return {cls.EI, cls.EI.upper(), cls.EI.capitalize()}

    @classmethod
    def poi_values(cls):
        return {cls.POI, cls.POI.upper(), cls.POI.capitalize()}

    @classmethod
    def is_ucb(cls, value):
        return value in cls.ucb_values()

    @classmethod
    def is_ei(cls, value):
        return value in cls.ei_values()

    @classmethod
    def is_poi(cls, value):
        return value in cls.poi_values()


class GaussianProcessesKernels(str, PEnum):
    RBF = "rbf"
    MATERN = "matern"

    @classmethod
    def rbf_value(cls):
        return {cls.RBF, cls.RBF.upper(), cls.RBF.capitalize()}

    @classmethod
    def matern_value(cls):
        return {cls.MATERN, cls.MATERN.upper(), cls.MATERN.capitalize()}

    @classmethod
    def is_rbf(cls, value):
        return value in cls.rbf_value()

    @classmethod
    def is_mattern(cls, value):
        return value in cls.matern_value()



from typing import Dict

from polyaxon._k8s import k8s_schemas


def get_custom_object(
    namespace: str,
    resource_name: str,
    kind: str,
    api_version: str,
    labels: Dict,
    annotations: Dict,
    custom_object: Dict,
) -> Dict:
    metadata = k8s_schemas.V1ObjectMeta(
        name=resource_name,
        labels=labels,
        annotations=annotations,
        namespace=namespace,
    )
    custom_object.update(
        {"kind": kind, "apiVersion": api_version, "metadata": metadata}
    )

    return custom_object



from polyaxon._compiler.lineage.artifacts_collector import (
    collect_lineage_artifacts_path,
)
from polyaxon._compiler.lineage.collector import resolve_artifacts_lineage
from polyaxon._compiler.lineage.io_collector import collect_io_artifacts



from typing import Optional

from polyaxon import settings
from polyaxon._client.run import RunClient
from polyaxon.exceptions import PolyaxonClientException, PolyaxonContainerException


def get_client_or_raise() -> Optional[RunClient]:
    if settings.CLIENT_CONFIG.no_api:
        return None

    try:
        return RunClient()
    except PolyaxonClientException as e:
        raise PolyaxonContainerException(e)



import os

from typing import Dict

from polyaxon._client.init import get_client_or_raise
from traceml.artifacts import V1ArtifactKind


def create_file_lineage(filepath: str, summary: Dict, kind: str):
    if not filepath:
        return
    filename = os.path.basename(filepath)
    if not kind:
        if "dockerfile" in filename.lower():
            kind = V1ArtifactKind.DOCKERFILE
        else:
            kind = V1ArtifactKind.FILE

    run_client = get_client_or_raise()
    if not run_client:
        return

    run_client.log_artifact_ref(
        path=filepath,
        kind=kind,
        name=filename,
        summary=summary,
        is_input=True,
    )



import asyncio
import traceback

from typing import Any, Optional, Tuple

from clipped.utils.tz import now
from clipped.utils.versions import clean_version_for_check
from clipped.utils.workers import async_exit_context, get_wait
from kubernetes.client.rest import ApiException
from urllib3.exceptions import HTTPError

from polyaxon import pkg, settings
from polyaxon._env_vars.getters import get_run_info
from polyaxon._runner.agent.base_agent import BaseAgent
from polyaxon._sdk.schemas.v1_agent import V1Agent
from polyaxon._sdk.schemas.v1_agent_state_response import V1AgentStateResponse
from polyaxon.exceptions import ApiException as SDKApiException
from polyaxon.exceptions import PolyaxonAgentError, PolyaxonConverterError
from polyaxon.logger import logger


class BaseAsyncAgent(BaseAgent):
    IS_ASYNC = True

    async def _enter(self):
        if not self.client._is_managed:
            return self
        print("Agent is starting.")
        await self.executor.refresh()
        try:
            agent = await self.client.get_info()
            self._check_status(agent)
            await self.sync()
            await self.client.log_agent_running()
            print("Agent is running.")
            return self
        except (ApiException, SDKApiException, HTTPError) as e:
            message = "Could not start the agent."
            if e.status == 404:
                reason = "Agent not found."
            elif e.status == 403:
                reason = "Agent is not approved yet or has invalid token."
            else:
                reason = "Error {}.".format(repr(e))
            await self.client.log_agent_failed(message="{} {}".format(message, reason))
            raise PolyaxonAgentError(message="{} {}".format(message, reason))
        except Exception as e:
            raise PolyaxonAgentError from e

    async def _exit(self):
        if not self.client._is_managed:
            return
        if not self._graceful_shutdown:
            await self.client.log_agent_warning()
        await asyncio.sleep(1)

    async def __aenter__(self):
        return await self._enter()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit()

    async def refresh_executor(self):
        if (
            now() - self._executor_refreshed_at
        ).total_seconds() > settings.AGENT_CONFIG.get_executor_refresh_interval():
            logger.debug("Refreshing executor ... ")
            await self.executor.refresh()
            self._executor_refreshed_at = now()

    async def sync(self):
        version_api = await self.executor.manager.get_version()
        return await self.client.sync_agent(
            agent=V1Agent(
                content=settings.AGENT_CONFIG.to_json(),
                version=clean_version_for_check(pkg.VERSION),
                version_api=version_api,
            ),
        )

    async def start(self):
        try:
            async with async_exit_context() as exit_event:
                index = 0
                timeout = get_wait(index, max_interval=self.max_interval)

                while True:
                    try:
                        await asyncio.wait_for(exit_event.wait(), timeout=timeout)
                        break  # If exit_event is set, we break out of the loop
                    except asyncio.TimeoutError:
                        index += 1
                        await self.refresh_executor()
                        if not self._default_auth:
                            await self.cron()
                        agent_state = await self.process()
                        if not agent_state:
                            logger.warning(
                                "Agent state is empty, waiting for next check."
                            )
                            index = self.max_interval
                            continue
                        self._check_status(agent_state)
                        if agent_state.state.full:
                            index = 2
                        self.ping()
                        timeout = get_wait(index, max_interval=self.max_interval)
                        logger.info("Sleeping for {} seconds".format(timeout))
        except Exception as e:
            print(e)
        finally:
            self.end()

    async def process(self, **kwargs) -> V1AgentStateResponse:
        try:
            agent_state = await self.client.get_state()
            if agent_state.compatible_updates:
                self.sync_compatible_updates(agent_state.compatible_updates)

            if agent_state:
                logger.info("Checking agent state.")
            else:
                logger.info("No state was found.")
                return V1AgentStateResponse.construct()

            state = agent_state.state
            if not state:
                return agent_state
            for run_data in state.schedules or []:
                await self.submit_run(run_data)
            for run_data in state.queued or []:
                await self.submit_run(run_data)
            for run_data in state.checks or []:
                await self.check_run(run_data)
            for run_data in state.stopping or []:
                await self.stop_run(run_data)
            for run_data in state.apply or []:
                await self.apply_run(run_data)
            for run_data in state.deleting or []:
                await self.delete_run(run_data)
            for run_data in state.hooks or []:
                await self.make_and_create_run(run_data)
            for run_data in state.watchdogs or []:
                await self.make_and_create_run(run_data)
            for run_data in state.tuners or []:
                await self.make_and_create_run(run_data, True)
            return agent_state
        except Exception as exc:
            logger.error(exc)
            return V1AgentStateResponse.construct()

    async def prepare_run_resource(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
    ) -> Optional[Any]:
        try:
            return await self.executor.convert(
                owner_name=owner_name,
                project_name=project_name,
                run_name=run_name,
                run_uuid=run_uuid,
                content=content,
                default_auth=self._default_auth,
                agent_content=self.content,
            )
        except PolyaxonConverterError as e:
            await self.client.log_run_failed(
                run_owner=owner_name,
                run_project=project_name,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed converting run manifest.\n",
            )
        except Exception as e:
            await self.client.log_run_failed(
                run_owner=owner_name,
                run_project=project_name,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed during compilation with unknown exception.\n",
            )
        return None

    async def submit_run(self, run_data: Tuple[str, str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = await self.prepare_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
        )
        if not resource:
            return

        namespace = None if len(run_data) < 5 else run_data[4]
        try:
            await self.executor.create(
                run_uuid=run_uuid,
                run_kind=run_data[1],
                resource=resource,
                namespace=namespace,
            )
        except ApiException as e:
            if e.status == 409:
                logger.info("Run already running, triggering an apply mechanism.")
                await self.apply_run(run_data=run_data)
            else:
                logger.info("Run submission error.")
                await self.client.log_run_failed(
                    run_owner=run_owner,
                    run_project=run_project,
                    run_uuid=run_uuid,
                    exc=e,
                )
        except Exception as e:
            await self.client.log_run_failed(
                run_owner=run_owner,
                run_project=run_project,
                run_uuid=run_uuid,
                exc=e,
            )

    async def make_and_create_run(
        self, run_data: Tuple[str, str, str, str, str], default_auth: bool = False
    ):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = await self.make_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
            default_auth=default_auth,
        )
        if not resource:
            return

        namepsace = None if len(run_data) < 5 else run_data[4]

        try:
            await self.executor.create(
                run_uuid=run_uuid,
                run_kind=run_data[1],
                resource=resource,
                namespace=namepsace,
            )
        except ApiException as e:
            if e.status == 409:
                logger.info("Run already running, triggering an apply mechanism.")
            else:
                logger.info("Run submission error.")
        except Exception as e:
            logger.info(
                "Run could not be cleaned. Agent failed converting run manifest: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )

    async def apply_run(self, run_data: Tuple[str, str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = await self.prepare_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
        )
        if not resource:
            return

        namespace = None if len(run_data) < 5 else run_data[4]

        try:
            await self.executor.apply(
                run_uuid=run_uuid,
                run_kind=run_data[1],
                resource=resource,
                namespace=namespace,
            )
            await self.client.log_run_running(
                run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
            )
        except Exception as e:
            await self.client.log_run_failed(
                run_owner=run_owner, run_project=run_project, run_uuid=run_uuid, exc=e
            )
            await self.clean_run(
                run_uuid=run_uuid, run_kind=run_data[1], namespace=namespace
            )

    async def check_run(self, run_data: Tuple[str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        namespace = None if len(run_data) < 3 else run_data[2]
        try:
            await self.executor.get(
                run_uuid=run_uuid, run_kind=run_data[1], namespace=namespace
            )
        except ApiException as e:
            if e.status == 404:
                logger.info(
                    "Run does not exist anymore, it could have been stopped or deleted."
                )
                await self.client.log_run_stopped(
                    run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
                )

    async def stop_run(self, run_data: Tuple[str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        namespace = None if len(run_data) < 3 else run_data[2]
        try:
            await self.executor.stop(
                run_uuid=run_uuid, run_kind=run_data[1], namespace=namespace
            )
        except ApiException as e:
            if e.status == 404:
                logger.info("Run does not exist anymore, it could have been stopped.")
                await self.client.log_run_stopped(
                    run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
                )
        except Exception as e:
            await self.client.log_run_failed(
                run_owner=run_owner,
                run_project=run_project,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed stopping run.\n",
            )

    async def delete_run(self, run_data: Tuple[str, str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        namespace = None if len(run_data) < 5 else run_data[4]
        if run_data[3]:
            await self.make_and_create_run(run_data)
        else:
            await self.clean_run(
                run_uuid=run_uuid, run_kind=run_data[1], namespace=namespace
            )

    async def clean_run(self, run_uuid: str, run_kind: str, namespace: str = None):
        try:
            await self.executor.clean(
                run_uuid=run_uuid, run_kind=run_kind, namespace=namespace
            )
            await self.executor.stop(
                run_uuid=run_uuid, run_kind=run_kind, namespace=namespace
            )
        except ApiException as e:
            if e.status == 404:
                logger.info("Run does not exist.")
        except Exception as e:
            logger.info(
                "Run could not be cleaned: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._schemas.lifecycle import V1StatusCondition


class V1EntityStatusBodyRequest(BaseAllowSchemaModel):
    owner: Optional[StrictStr]
    project: Optional[StrictStr]
    uuid: Optional[UUIDStr]
    condition: Optional[V1StatusCondition]
    force: Optional[bool]



from polyaxon._constants.globals import UNKNOWN


class EventTypes:
    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"
    ERROR = "ERROR"


class PodConditions:
    READY = "Ready"
    INITIALIZED = "Initialized"
    SCHEDULED = "PodScheduled"
    UNSCHEDULABLE = "Unschedulable"

    VALUES = [READY, INITIALIZED, SCHEDULED]


class PodLifeCycle:
    CONTAINER_CREATING = "ContainerCreating"
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    UNKNOWN = UNKNOWN

    DONE_STATUS = [FAILED, SUCCEEDED]



from typing import Union
from typing_extensions import Annotated

from clipped.compact.pydantic import Field

from polyaxon._flow.run.cleaner import V1CleanerJob
from polyaxon._flow.run.dag import V1Dag
from polyaxon._flow.run.dask import V1DaskJob, V1DaskReplica
from polyaxon._flow.run.enums import (
    V1CloningKind,
    V1PipelineKind,
    V1RunEdgeKind,
    V1RunKind,
    V1RunPending,
)
from polyaxon._flow.run.job import V1Job
from polyaxon._flow.run.kubeflow.clean_pod_policy import V1CleanPodPolicy
from polyaxon._flow.run.kubeflow.mpi_job import V1MPIJob
from polyaxon._flow.run.kubeflow.mx_job import MXJobMode, V1MXJob
from polyaxon._flow.run.kubeflow.paddle_job import V1PaddleElasticPolicy, V1PaddleJob
from polyaxon._flow.run.kubeflow.pytorch_job import V1PytorchElasticPolicy, V1PytorchJob
from polyaxon._flow.run.kubeflow.replica import V1KFReplica
from polyaxon._flow.run.kubeflow.scheduling_policy import V1SchedulingPolicy
from polyaxon._flow.run.kubeflow.tf_job import V1TFJob
from polyaxon._flow.run.kubeflow.xgboost_job import V1XGBoostJob
from polyaxon._flow.run.notifier import V1NotifierJob
from polyaxon._flow.run.patch import validate_run_patch
from polyaxon._flow.run.ray import V1RayJob, V1RayReplica
from polyaxon._flow.run.resources import V1RunResources
from polyaxon._flow.run.service import V1Service
from polyaxon._flow.run.tuner import V1TunerJob

V1Runtime = Annotated[
    Union[
        V1Job,
        V1Service,
        V1Dag,
        V1MPIJob,
        V1PytorchJob,
        V1TFJob,
        V1MXJob,
        V1PaddleJob,
        V1XGBoostJob,
        V1DaskJob,
        V1RayJob,
        V1NotifierJob,
        V1CleanerJob,
        V1TunerJob,
    ],
    Field(discriminator="kind"),
]


class RunMixin:
    def get_run_kind(self):
        raise NotImplementedError

    @property
    def is_job_run(self):
        return self.get_run_kind() == V1RunKind.JOB

    @property
    def is_service_run(self):
        return self.get_run_kind() == V1RunKind.SERVICE

    @property
    def is_mpi_job_run(self):
        return self.get_run_kind() == V1RunKind.MPIJOB

    @property
    def is_paddle_job_run(self):
        return self.get_run_kind() == V1RunKind.PADDLEJOB

    @property
    def is_pytorch_job_run(self):
        return self.get_run_kind() == V1RunKind.PYTORCHJOB

    @property
    def is_tf_job_run(self):
        return self.get_run_kind() == V1RunKind.TFJOB

    @property
    def is_mx_job_run(self):
        return self.get_run_kind() == V1RunKind.MXJOB

    @property
    def is_xgb_job_run(self):
        return self.get_run_kind() == V1RunKind.XGBJOB

    @property
    def is_ray_job_run(self):
        return self.get_run_kind() == V1RunKind.RAYJOB

    @property
    def is_dask_job_run(self):
        return self.get_run_kind() == V1RunKind.DASKJOB

    @property
    def is_dag_run(self):
        return self.get_run_kind() == V1RunKind.DAG

    @property
    def is_schedule_run(self):
        return self.get_run_kind() == V1RunKind.SCHEDULE

    @property
    def is_notifier_run(self):
        return self.get_run_kind() == V1RunKind.NOTIFIER

    @property
    def is_cleaner_run(self):
        return self.get_run_kind() == V1RunKind.CLEANER

    @property
    def is_tuner_run(self):
        return self.get_run_kind() == V1RunKind.TUNER

    @property
    def is_watchdog(self):
        return self.get_run_kind() == V1RunKind.WATCHDOG

    @property
    def is_distributed_run(self):
        return (
            self.is_mpi_job_run
            or self.is_pytorch_job_run
            or self.is_tf_job_run
            or self.is_mx_job_run
            or self.is_xgb_job_run
            or self.is_ray_job_run
            or self.is_dask_job_run
        )



from clipped.utils.validation import validate_tags

from polyaxon._schemas.types import V1TensorboardType

TENSORBOARD_INIT_COMMAND = ["polyaxon", "initializer", "tensorboard"]


def get_tensorboard_args(
    tb_args: V1TensorboardType, context_from: str, context_to: str, connection_kind: str
):
    args = [
        "--context-from={}".format(context_from),
        "--context-to={}".format(context_to),
        "--connection-kind={}".format(connection_kind),
    ]
    if tb_args.port:
        args.append("--port={}".format(tb_args.port))
    if tb_args.uuids:
        uuids = validate_tags(tb_args.uuids, validate_yaml=True)
        args.append("--uuids={}".format(",".join(uuids)))
    if tb_args.use_names:
        args.append("--use-names")
    if tb_args.path_prefix:
        args.append("--path-prefix={}".format(tb_args.path_prefix)),
    if tb_args.plugins:
        plugins = validate_tags(tb_args.plugins, validate_yaml=True)
        args.append("--plugins={}".format(",".join(plugins)))

    return args



from polyaxon.exceptions import PolyaxonSchemaError


def get_queue_info(queue: str):
    if not queue:
        raise PolyaxonSchemaError("Received an invalid queue {}".format(queue))

    parts = queue.replace(".", "/").split("/")
    agent = None
    queue_name = queue
    if len(parts) == 2:
        agent, queue_name = parts
    elif len(parts) > 2:
        raise PolyaxonSchemaError(
            "Please provide a valid queue. "
            "The queue name should be: queue-name to use the default agent or agent-name/queue."
        )

    return agent, queue_name



from typing import Dict

from polyaxon._compiler.contexts.base import BaseContextsManager
from polyaxon._connections import V1Connection
from polyaxon._flow import V1CompiledOperation


class JobContextsManager(BaseContextsManager):
    @classmethod
    def resolve(
        cls,
        namespace: str,
        owner_name: str,
        project_name: str,
        run_uuid: str,
        contexts: Dict,
        compiled_operation: V1CompiledOperation,
        connection_by_names: Dict[str, V1Connection],
    ) -> Dict:
        return cls._resolver_replica(
            contexts=contexts,
            init=compiled_operation.run.init,
            connections=compiled_operation.run.connections,
            connection_by_names=connection_by_names,
        )



import base64

from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1LogHandler(BaseAllowSchemaModel):
    _IDENTIFIER = "log_handler"

    dsn: Optional[StrictStr]
    environment: Optional[StrictStr]

    @property
    def decoded_dsn(self):
        if self.dsn:
            return base64.b64decode(self.dsn.encode("utf-8")).decode("utf-8")



from typing import Optional

from clipped.compact.pydantic import StrictStr

from polyaxon._flow.optimization.enums import V1Optimization, V1ResourceType
from polyaxon._schemas.base import BaseSchemaModel


class V1OptimizationMetric(BaseSchemaModel):
    _IDENTIFIER = "optimization_metric"

    name: StrictStr
    optimization: Optional[V1Optimization]

    def get_for_sort(self):
        if self.optimization == V1Optimization.MINIMIZE:
            return self.name
        return "-{}".format(self.name)


class V1OptimizationResource(BaseSchemaModel):
    _IDENTIFIER = "optimization_resource"

    name: StrictStr
    type: Optional[V1ResourceType]

    def cast_value(self, value):
        if V1ResourceType.is_int(self.type):
            return int(value)
        if V1ResourceType.is_float(self.type):
            return float(value)
        return value



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1PasswordChange(BaseAllowSchemaModel):
    old_password: Optional[StrictStr]
    new_password1: Optional[StrictStr]
    new_password2: Optional[StrictStr]



from typing import Dict, List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1ArtifactTree(BaseAllowSchemaModel):
    files: Optional[Dict[str, int]]
    dirs: Optional[List[StrictStr]]
    is_done: Optional[bool]



from typing import Dict, Optional, Type

from clipped.utils.lists import to_list

from polyaxon._flow import (
    V1IO,
    V1CompiledOperation,
    V1Component,
    V1Operation,
    V1Param,
    validate_run_patch,
)
from polyaxon._flow.operations.operation import PartialV1Operation
from polyaxon._polyaxonfile.specs import kinds
from polyaxon._polyaxonfile.specs.base import BaseSpecification
from polyaxon.exceptions import PolyaxonSchemaError


class OperationSpecification(BaseSpecification):
    """The polyaxonfile specification for operations."""

    _SPEC_KIND = kinds.OPERATION

    CONFIG: Type[V1Operation] = V1Operation
    PARTIAL_CONFIG: Type[PartialV1Operation] = PartialV1Operation

    @classmethod
    def compile_operation(
        cls,
        config: V1Operation,
        override: Optional[Dict] = None,
        use_override_patch_strategy: bool = False,
    ) -> V1CompiledOperation:
        preset_patch_strategy = None
        if override:
            preset = OperationSpecification.read(override, is_preset=True)
            if use_override_patch_strategy and preset.patch_strategy:
                preset_patch_strategy = preset.patch_strategy

            config = config.patch(preset, preset.patch_strategy)
        # Patch run
        component = config.component  # type: V1Component
        if not component:
            raise PolyaxonSchemaError(
                "Compile operation received an invalid configuration: "
                "the component is missing. "
                "Please make sure that the polyaxonfile was correctly resolved "
                "before to calling this operation."
            )
        if config.run_patch:
            patch_strategy = (
                preset_patch_strategy
                if use_override_patch_strategy and preset_patch_strategy is not None
                else config.patch_strategy
            )
            component.run = component.run.patch(
                validate_run_patch(config.run_patch, component.run.kind),
                strategy=patch_strategy,
            )

        contexts = []

        def get_context_io(c_name: str, c_io: V1Param, is_list=None):
            if not c_io.context_only:
                return

            contexts.append(
                V1IO.construct(
                    name=c_name,
                    to_init=c_io.to_init,
                    to_env=c_io.to_env,
                    connection=c_io.connection,
                    is_list=is_list,
                )
            )

        # Collect contexts io form params
        for p in config.params or {}:
            get_context_io(c_name=p, c_io=config.params[p])

        # Collect contexts io form joins
        for j in config.joins or []:
            for p in j.params or {}:
                get_context_io(c_name=p, c_io=j.params[p], is_list=True)

        patch_keys = {
            "name",
            "cost",
            "description",
            "contexts",
            "tags",
            "is_approved",
            "presets",
            "queue",
            "namespace",
            "cache",
            "build",
            "hooks",
            "events",
            "plugins",
            "termination",
            "matrix",
            "joins",
            "schedule",
            "dependencies",
            "trigger",
            "conditions",
            "skip_on_upstream_skip",
        }
        patch_keys = patch_keys.intersection(config.__fields_set__)
        patch_data = {k: getattr(config, k) for k in patch_keys}
        patch_compiled = V1CompiledOperation.construct(contexts=contexts, **patch_data)

        values = [
            {cls.VERSION: config.version},
            component.to_dict(),
            {cls.KIND: kinds.COMPILED_OPERATION},
        ]
        compiled = V1CompiledOperation.read(values)  # type: V1CompiledOperation
        return compiled.patch(patch_compiled, strategy=config.patch_strategy)

    @classmethod
    def read(cls, values, partial: bool = False, is_preset: bool = False):
        if is_preset:
            if isinstance(values, cls.CONFIG):
                values.is_preset = True
                return values
            elif isinstance(values, Dict):
                values[cls.IS_PRESET] = True
            else:
                values = to_list(values)
                values = [{cls.IS_PRESET: True}] + values

        return super().read(values, partial=partial or is_preset)



from typing import Dict, List, Optional, Union

from clipped.compact.pydantic import Field, StrictStr, validator
from clipped.types.ref_or_obj import BoolOrRef, RefField

from polyaxon._flow.params import V1Param
from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._schemas.lifecycle import V1Statuses


class V1Hook(BaseSchemaModel):
    """You can configure Polyaxon to send notifications and webhooks to users and systems
    when operations reaches a final state,
    or trigger any logic that is tightly coupled with a class of operations.

    Hooks allow you to build or set up integrations and dependencies
    based on final events generated by Polyaxon.
    External systems can subscribe or provide handling for certain events.
    When one of those events is triggered, Polyaxon will trigger the components defined
    in the hooks sections to perform a post-done logic, like sending an HTTP request
    payload to the webhook's configured URL.

    In addition to the main use-case of hooks, i.e. notifications,
    you can also use hooks to trigger a post-success logic that does
    not have to run on the same container, does not require specific accelerators,
    or is not specific to a single operation.
    For instance, you can build a custom hook to:
      * run evaluation after a training operation succeeds.
      * perform data or model checks.

    Polyaxon can trigger hooks when a run reaches a final status:

     * succeeded
     * failed
     * stopped
     * done (any final state)

    You can additionally provide a set of `conditions`
    to perform additional checks before triggering the logic, for instance,
    in addition to the success status, you can restrict the hook to only trigger
    if a metric has reached a certain value: `conditions: {{ loss > 0.9 }}`.

    You can resolve any context information from the main operation inside hooks,
    like params, globals, ...

    Args:
        trigger: str
        connection: str
        hub_ref: str
        conditions: str, optional
        queue: str, optional
        namespace: str, optional
        presets: List[str], optional
        disableDefaults: bool, optional
        params: Dict[str, [V1Param](/docs/core/specification/params/)], optional

    ## YAML usage

    ```yaml
    >>> hook:
    >>>   trigger: failed
    >>>   connection: slack-connection
    >>>   hubRef: slack
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Statuses, V1Hook
    >>> hook = V1Hook(
    >>>     trigger=V1Statuses.FAILED,
    >>>     hub_ref="slack",
    >>>     connection="slack-connection",
    >>> )
    ```

    ## Fields

    ### connection

    The connection to notify, this [connection](/docs/setup/connections/)
    must be configured at deployment time to be used here by referencing the name.

    ```yaml
    >>> hook:
    >>>   connection: some-connection
    ```

    ### trigger

    The trigger represents the status condition to check before sending the notification.

    ```yaml
    >>> hook:
    >>>   trigger: succeeded
    ```

    In this example, the notification will be sent if the run succeeds.

    ### hubRef

    Polyaxon provides a [Component Hub](/docs/management/component-hub/)
    for hosting versioned components with an access control system to improve
    the productivity of your team.

    To trigger a hook based on a component hosted on Polyaxon Component Hub, you can use `hubRef`

    ```yaml
    >>> hook:
    >>>   hubRef: slack
    ...
    ```

    Or custom hook component

    ```yaml
    >>> hook:
    >>>   hubRef: my-component:dev
    ...
    ```

    ### conditions

    After the main operation is done, conditions take advantage of all context values
    resolved from the main operation, including outputs, to decide if the hook can be started.


    ```yaml
    >>>   conditions: '{{ some-io-param == "some-value" }}'
    ```

    In the example above, the hook will only run if a param is passed, or an output is logged and
    is equal to "some-value".

    ### presets

    The [presets](/docs/management/organizations/presets/) to use for the hook operation,
    if provided, it will override the component's presets otherwise
    the presets of the component will be used if available.

    ```yaml
    >>> hook:
    >>>   presets: [test]
    ```

    ### queue

    The [queue](/docs/core/scheduling-strategies/queues/) to use.
    If not provided, the default queue will be used.

    ```yaml
    >>> hook:
    >>>   queue: agent-name/queue-name
    ```

    If the agent name is not specified, Polyaxon will resolve the name of the queue
    based on the default agent.

    ```yaml
    >>> hook:
    >>>   queue: queue-name
    ```

    ### namespace

    > **Note**: Please note that this field is only available in some commercial editions.

    The namespace to use, if not provided, it will default to the agent's namespace.

    ```yaml
    >>> hook:
    >>>   namespace: polyaxon
    ```

    ### disableDefaults

    One major difference between hooks and normal operations,
    is that hooks will be initialized automatically with:
    `uuid`, `kind`, `name`, `inputs`, `outputs`, `status`, and `condition` of the main operation as
    context only params, to reduce the boilerplate and the need to request usual information
    required for most notification operations.

    If you do not need these context values or if you decide to request params manually,
    you can set this field to `false`.

    ```yaml
    >>> hook:
    >>>   disableDefaults: true
    >>>   ...
    ```

    ### params

    The [params](/docs/core/specification/params/) to pass if the handler requires extra params,
    they will be validated against the inputs/outputs.
    If a parameter is passed and the component does not define a corresponding inputs/outputs,
    a validation error will be raised unless the param has the `contextOnly` flag enabled.

    ```yaml
    >>> hook:
    >>>   params:
    >>>     param1: {value: 1.1}
    >>>     param2: {value: test}
    >>>   ...
    ```
    """

    _IDENTIFIER = "hook"

    hub_ref: StrictStr = Field(alias="hubRef")
    connection: Optional[StrictStr]
    trigger: Optional[V1Statuses]
    conditions: Optional[StrictStr]
    presets: Optional[List[StrictStr]]
    queue: Optional[StrictStr]
    namespace: Optional[StrictStr]
    params: Optional[Union[Dict[str, V1Param], RefField]]
    disable_defaults: Optional[BoolOrRef] = Field(alias="disableDefaults")

    @validator("trigger")
    def validate_trigger(cls, v):
        if v and v not in V1Statuses.get_allowable_hook_values():
            raise ValueError(
                "Trigger must be one of {}".format(
                    V1Statuses.get_allowable_hook_values()
                )
            )
        return v

    def check_trigger_for_status(self, status: V1Statuses) -> bool:
        if not self.trigger or self.trigger == V1Statuses.DONE:
            return True

        return status == self.trigger



from clipped.utils.enums import PEnum


class V1ResourceType(str, PEnum):
    INT = "int"
    FLOAT = "float"

    @classmethod
    def init_values(cls):
        return {cls.INT, cls.INT.upper(), cls.INT.capitalize()}

    @classmethod
    def float_values(cls):
        return {cls.FLOAT, cls.FLOAT.upper(), cls.FLOAT.capitalize()}

    @classmethod
    def values(cls):
        return cls.init_values() | cls.float_values()

    @classmethod
    def is_int(cls, value):
        return value in cls.init_values()

    @classmethod
    def is_float(cls, value):
        return value in cls.float_values()


class V1Optimization(str, PEnum):
    MAXIMIZE = "maximize"
    MINIMIZE = "minimize"

    @classmethod
    def maximize_values(cls):
        return [cls.MAXIMIZE, cls.MAXIMIZE.upper(), cls.MAXIMIZE.capitalize()]

    @classmethod
    def minimize_values(cls):
        return [cls.MINIMIZE, cls.MINIMIZE.upper(), cls.MINIMIZE.capitalize()]

    @classmethod
    def maximize(cls, value):
        return value in cls.maximize_values()

    @classmethod
    def minimize(cls, value):
        return value in cls.minimize_values()



from polyaxon._env_vars.getters.agent import get_agent_info, get_artifacts_store_name
from polyaxon._env_vars.getters.owner_entity import resolve_entity_info
from polyaxon._env_vars.getters.project import (
    get_project_error_message,
    get_project_or_local,
)
from polyaxon._env_vars.getters.queue import get_queue_info
from polyaxon._env_vars.getters.run import (
    get_collect_artifacts,
    get_collect_resources,
    get_log_level,
    get_project_run_or_local,
    get_run_info,
    get_run_or_local,
)
from polyaxon._env_vars.getters.user import get_local_owner
from polyaxon._env_vars.getters.versioned_entity import (
    get_component_info,
    get_model_info,
    get_versioned_entity_info,
)
